/*
 *
 * Copyright (c) 2001
 *	Politecnico di Torino.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code
 * distributions retain the above copyright notice and this paragraph
 * in its entirety, (2) distributions including binary code include
 * the above copyright notice and this paragraph in its entirety in
 * the documentation or other materials provided with the
 * distribution, and (3) all advertising materials mentioning features
 * or use of this software display the following acknowledgment:
 * ``This product includes software developed by the Politecnico di Torino
 * Telecomunication Networking Group.''
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Author:	Marco Mellia, Andrea Carpani
 * 		Telecomunication Networks Group
 * 		Politecnico di Torino
 * 		Torino, Italy
 *              http://http://www.tlc-networks.polito.it/index.html
 *		mellia@mail.tlc.polito.it
 *
 * Tstat is deeply based on TCPTRACE. The authors would like to thank
 * Shawn Ostermann for the development of TCPTRACE.
 * The original TCPTRACE copyright can be found in tcptrace.copyright file.
 *
*/


#include "tstat.h"


/* provided globals  */
extern FILE *fp_logc;
extern FILE *fp_lognc;
extern Bool is_stdin;
extern Bool printticks;
extern struct adx *adx_index[MAX_ADX_SLOTS];
extern unsigned long int fcount;

/* tcp database stats */
long int tcp_packet_count = 0;
int search_count = 0;

int num_tcp_pairs = 0;		/* how many pairs we've allocated */
tcp_pair **ttp = NULL;		/* array of pointers to allocated pairs */
struct tp_list_elem *tp_list_start = NULL;	/* starting point of the linked list */
struct tp_list_elem *tp_list_curr = NULL;	/* current insert point of the linked list */
u_long tcp_trace_count = 0;


/* local routine definitions */
static tcp_pair **NewTTP_2 (struct ip *, struct tcphdr *);
static ptp_snap **FindTTP (struct ip *, struct tcphdr *, int *);


/* return elapsed time in microseconds */
/* (time2 - time1) */
double
elapsed (struct timeval time1, struct timeval time2)
{
  struct timeval etime;

  /*sanity check, some of the files have packets out of order */
  if (tv_lt (time2, time1))
    {
      return (0.0);
    }

  etime = time2;
  tv_sub (&etime, time1);

  return (time2double (etime));
}



/* subtract the rhs from the lhs, result in lhs */
void
tv_sub (struct timeval *plhs, struct timeval rhs)
{
  /* sanity check, lhs MUST BE more than rhs */
  if (tv_lt (*plhs, rhs))
    {
      fprintf (stderr, "tvsub(%s,", ts2ascii (plhs));
      fprintf (stderr, "%s) bad timestamp order!\n", ts2ascii (&rhs));
/* 	exit(-1); */
      plhs->tv_sec = plhs->tv_usec = 0;
      return;
    }

  if (plhs->tv_usec >= rhs.tv_usec)
    {
      plhs->tv_usec -= rhs.tv_usec;
    }
  else if (plhs->tv_usec < rhs.tv_usec)
    {
      plhs->tv_usec += US_PER_SEC - rhs.tv_usec;
      plhs->tv_sec -= 1;
    }
  plhs->tv_sec -= rhs.tv_sec;
}


/* add the RHS to the LHS, answer in *plhs */
void
tv_add (struct timeval *plhs, struct timeval rhs)
{
  plhs->tv_sec += rhs.tv_sec;
  plhs->tv_usec += rhs.tv_usec;

  if (plhs->tv_usec >= US_PER_SEC)
    {
      plhs->tv_usec -= US_PER_SEC;
      plhs->tv_sec += 1;
    }
}


/* are the 2 times the same? */
Bool
tv_same (struct timeval lhs, struct timeval rhs)
{
  return ((lhs.tv_sec == rhs.tv_sec) && (lhs.tv_usec == rhs.tv_usec));
}


/*  1: lhs >  rhs */
/*  0: lhs == rhs */
/* -1: lhs <  rhs */
int
tv_cmp (struct timeval lhs, struct timeval rhs)
{
  if (lhs.tv_sec > rhs.tv_sec)
    {
      return (1);
    }

  if (lhs.tv_sec < rhs.tv_sec)
    {
      return (-1);
    }

  /* ... else, seconds are the same */
  if (lhs.tv_usec > rhs.tv_usec)
    return (1);
  else if (lhs.tv_usec == rhs.tv_usec)
    return (0);
  else
    return (-1);
}



/* copy the IP addresses and port numbers into an addrblock structure	*/
/* in addition to copying the address, we also create a HASH value	*/
/* which is based on BOTH IP addresses and port numbers.  It allows	*/
/* faster comparisons most of the time					*/
void
CopyAddr (tcp_pair_addrblock * ptpa,
	  struct ip *pip, portnum port1, portnum port2)
{
  ptpa->a_port = port1;
  ptpa->b_port = port2;

  if (PIP_ISV4 (pip))
    {				/* V4 */
      IP_COPYADDR (&ptpa->a_address, *IPV4ADDR2ADDR (&pip->ip_src));
      IP_COPYADDR (&ptpa->b_address, *IPV4ADDR2ADDR (&pip->ip_dst));
      /* fill in the hashed address */
      ptpa->hash = ptpa->a_address.un.ip4.s_addr
	+ ptpa->b_address.un.ip4.s_addr + ptpa->a_port + ptpa->b_port;
    }
#ifdef SUPPORT_IPV6
  else
    {				/* V6 */
      int i;
      struct ipv6 *pip6 = (struct ipv6 *) pip;
      IP_COPYADDR (&ptpa->a_address, *IPV6ADDR2ADDR (&pip6->ip6_saddr));
      IP_COPYADDR (&ptpa->b_address, *IPV6ADDR2ADDR (&pip6->ip6_daddr));
      /* fill in the hashed address */
      ptpa->hash = ptpa->a_port + ptpa->b_port;
      for (i = 0; i < 16; ++i)
	{
	  ptpa->hash += ptpa->a_address.un.ip6.s6_addr[i];
	  ptpa->hash += ptpa->b_address.un.ip6.s6_addr[i];
	}
    }
#endif
}



int
WhichDir (tcp_pair_addrblock * ptpa1, tcp_pair_addrblock * ptpa2)
{

#ifdef BROKEN_COMPILER
  /* sorry for the ugly nested 'if', but a 4-way conjunction broke my */
  /* Optimizer (under 'gcc version cygnus-2.0.2')                     */

  /* same as first packet */
  if (IP_SAMEADDR (ptpa1->a_address, ptpa2->a_address))
    if (IP_SAMEADDR (ptpa1->b_address, ptpa2->b_address))
      if ((ptpa1->a_port == ptpa2->a_port))
	if ((ptpa1->b_port == ptpa2->b_port))
	  return (A2B);

  /* reverse of first packet */
  if (IP_SAMEADDR (ptpa1->a_address, ptpa2->b_address))
    if (IP_SAMEADDR (ptpa1->b_address, ptpa2->a_address))
      if ((ptpa1->a_port == ptpa2->b_port))
	if ((ptpa1->b_port == ptpa2->a_port))
	  return (B2A);
#else /* BROKEN_COMPILER */
  /* same as first packet */
  if (IP_SAMEADDR (ptpa1->a_address, ptpa2->a_address) &&
      IP_SAMEADDR (ptpa1->b_address, ptpa2->b_address) &&
      (ptpa1->a_port == ptpa2->a_port) && (ptpa1->b_port == ptpa2->b_port))
    return (A2B);

  /* reverse of first packet */
  if (IP_SAMEADDR (ptpa1->a_address, ptpa2->b_address) &&
      IP_SAMEADDR (ptpa1->b_address, ptpa2->a_address) &&
      (ptpa1->a_port == ptpa2->b_port) && (ptpa1->b_port == ptpa2->a_port))
    return (B2A);
#endif /* BROKEN_COMPILER */

  /* different connection */
  return (0);
}



int
SameConn (tcp_pair_addrblock * ptpa1, tcp_pair_addrblock * ptpa2, int *pdir)
{
  /* if the hash values are different, they can't be the same */
  if (ptpa1->hash != ptpa2->hash)
    return (0);

  /* OK, they hash the same, are they REALLY the same function */
  *pdir = WhichDir (ptpa1, ptpa2);
  return (*pdir != 0);
}

extern Bool warn_MAX_;
static tcp_pair **
NewTTP_2 (struct ip *pip, struct tcphdr *ptcp)
{
  tcp_pair *ptp;
  int old_new_tcp_pairs = num_tcp_pairs;
  int steps=0;

  /* look for the next eventually available free block */
  num_tcp_pairs++;
  num_tcp_pairs = num_tcp_pairs % MAX_TCP_PAIRS;
  /* make a new one, if possible */
  while ((num_tcp_pairs != old_new_tcp_pairs) && (ttp[num_tcp_pairs] != NULL) && (steps < LIST_SEARCH_DEPT))
    {
      steps++;
      /* look for the next one */
//         printf("%d %d\n", num_tcp_pairs, old_new_tcp_pairs);
      num_tcp_pairs++;
      num_tcp_pairs = num_tcp_pairs % MAX_TCP_PAIRS;
    }
  if (ttp[num_tcp_pairs] != NULL)
    {
      if (warn_MAX_)
	{
	  printf
	    ("\nooopsss: number of simultaneous connection opened is greater then the maximum supported number!\n");
	  printf
	    ("you have to rebuild the source with a larger MAX_TCP_PAIRS defined!\n");
	}
      warn_MAX_ = FALSE;
      return (NULL);
    }

  /* create a new TCP pair record and remember where you put it */
  ptp = ttp[num_tcp_pairs] = tp_alloc ();

  /* grab the address from this packet */
  CopyAddr (&ptp->addr_pair,
	    pip, ntohs (ptcp->th_sport), ntohs (ptcp->th_dport));

  ptp->a2b.time.tv_sec = -1;
  ptp->b2a.time.tv_sec = -1;
  /* a.c */
  ptp->b2a.closed = FALSE;
  ptp->a2b.closed = FALSE;

  ptp->a2b.ptp = ptp;
  ptp->b2a.ptp = ptp;
  ptp->a2b.ptwin = &ptp->b2a;
  ptp->b2a.ptwin = &ptp->a2b;

  ptp->internal_src = internal_src;
  ptp->internal_dst = internal_dst;
  return (&ttp[num_tcp_pairs]);
}

static ptp_snap *
NewPTPH_2 (void)
{
  return (ptph_alloc ());
}

void
acstrncpy (char *d, char *s, int n)
{
  int i;

  i = 1;
  while ((*d++ = *s++) != '\0')
    {
      if (i++ == n)
	return;
    }
}


/* connection records are stored in a hash table.  Buckets are linked	*/
/* lists sorted by most recent access.					*/

ptp_snap *ptp_hashtable[HASH_TABLE_SIZE] = { NULL };

static ptp_snap **
FindTTP (struct ip *pip, struct tcphdr *ptcp, int *pdir)
{
  ptp_snap **pptph_head = NULL;
  ptp_snap *ptph;
  ptp_snap *ptph_last;
  static tcp_pair **temp_ttp;

  tcp_pair_addrblock tp_in;
  int dir;
  hash hval;

  /* grab the address from this packet */
  CopyAddr (&tp_in, pip, ntohs (ptcp->th_sport), ntohs (ptcp->th_dport));

  /* grab the hash value (already computed by CopyAddr) */
  hval = tp_in.hash % HASH_TABLE_SIZE;

  ptph_last = NULL;
  pptph_head = &ptp_hashtable[hval];


  for (ptph = *pptph_head; ptph; ptph = ptph->next)
    {
      ++search_count;

      if (SameConn (&tp_in, &ptph->addr_pair, &dir))
	{
	  /* OK, this looks good, suck it into memory */
	  tcp_pair *ptp = ptph->ptp;
	  tcb *thisdir;
	  tcb *otherdir;


	  /* figure out which direction this packet is going */
	  if (dir == A2B)
	    {
	      thisdir = &ptp->a2b;
	      otherdir = &ptp->b2a;
	    }
	  else
	    {
	      thisdir = &ptp->b2a;
	      otherdir = &ptp->a2b;
	    }

	  /* move to head of access list (unless already there) */
	  if (ptph != *pptph_head)
	    {
	      ptph_last->next = ptph->next;	/* unlink */
	      ptph->next = *pptph_head;	/* move to head */
	      *pptph_head = ptph;
	    }
	  *pdir = dir;
	  return (pptph_head);
	}
      ptph_last = ptph;
    }

  /* Didn't find it, make a new one, if possible */

  if (!SYN_SET (ptcp))
    {
      /* the new connection must begin with a SYN */
      if (debug > 1)
	{
	  printf
	    ("** trash TCP packet: it does not belong to any know flows\n");
	}
      not_id_p++;
      return (NULL);
    }

  /* ogni nuovo flusso controllo se son passati piu' di MAX_TIME_STEP secondi dall'inizio */
  if (elapsed (last_cleaned, current_time) > IDLE_TIME)
    {
      trace_done_periodic ();
      last_cleaned = current_time;
    }
  if (elapsed (last_time_step, current_time) > MAX_TIME_STEP)
    {
      /* metto in curr_data_dir la base piu' un decimale per lo "step" */
      sprintf (curr_data_dir, "%s/%03d", basename, step);
      print_all_histo ();
      clear_all_histo ();
      print_adx (adx_index);
      step++;
      last_time_step = current_time;
      /* controllo quelli ormai "in disuso" */
      if (step >= 4 && is_stdin)
	{
	  create_new_outfiles (NULL);
	  step = 0;
	}
    }

  fcount++;
  //printf("%lu\r",tot_conn);
  temp_ttp = NewTTP_2 (pip, ptcp);
  if (temp_ttp == NULL)		/* not enough memory to store the new flow */
    {
      /* the new connection must begin with a SYN */
      if (debug)
	{
	  printf ("** out of memory - considering a not_id_p\n");
	}
      not_id_p++;
      return (NULL);
    }
  tot_conn++;
  ptph = NewPTPH_2 ();
  ptph->ttp_ptr = temp_ttp;
  ptph->ptp = *(ptph->ttp_ptr);

  ptph->addr_pair = ptph->ptp->addr_pair;

  /* put at the head of the access list */
  ptph->next = *pptph_head;
  *pptph_head = ptph;

  *pdir = A2B;

  /* esco con il ptph che ho appena fatto */
  return (pptph_head);
}

tcp_pair *
dotrace_2 (struct ip * pip, struct tcphdr * ptcp, void *plast)
{
  struct tcp_options *ptcpo;
  tcp_pair *ptp_save;
  ptp_snap **ptph_ptr;
  ptp_snap *ptph_save;
  ptp_snap *ptph_tmp;
  int tcp_length;
  int tcp_data_length;
  u_long start;
  u_long end;
  tcb *thisdir;
  tcb *otherdir;
  tcp_pair tp_in;
  int dir;
  Bool retrans;
  Bool ecn_ce = FALSE;
  Bool ecn_echo = FALSE;
  Bool cwr = FALSE;
  int retrans_num_bytes;
  Bool out_order;		/* out of order */
  u_short th_sport;		/* source port */
  u_short th_dport;		/* destination port */
  tcp_seq th_seq;		/* sequence number */
  tcp_seq th_ack;		/* acknowledgement number */
  u_short th_win;		/* window */
  u_long eff_win;		/* window after scaling */
  short ip_len;			/* total length */
  enum t_ack ack_type = NORMAL;	/* how should we draw the ACK */
  seqnum old_this_windowend;	/* for graphing */

  /* make sure we have enough of the packet */
  if ((unsigned) ptcp + sizeof (struct tcphdr) - 1 > (unsigned) plast)
    {
      if (warn_printtrunc)
	fprintf (stderr,
		 "TCP packet %lu truncated too short to trace, ignored\n",
		 pnum);
      ++ctrunc;
      return (NULL);
    }


  /* convert interesting fields to local byte order */
  th_seq = ntohl (ptcp->th_seq);
  th_ack = ntohl (ptcp->th_ack);
  th_sport = ntohs (ptcp->th_sport);
  th_dport = ntohs (ptcp->th_dport);
  th_win = ntohs (ptcp->th_win);
  ip_len = gethdrlength (pip, plast) + getpayloadlength (pip, plast);

  /* make sure this is one of the connections we want */
  ptph_ptr = FindTTP (pip, ptcp, &dir);

  /* if the connection is not to be analyzed return a NULL */
  if (ptph_ptr == NULL)
    {
      return (NULL);
    }

  ptph_save = (*ptph_ptr);
  ptp_save = ptph_save->ptp;

  ++tcp_packet_count;

  if (ptp_save == NULL)
    {
      return (NULL);
    }

  ++tcp_trace_count;

  /* .a.c. tcp packet level stats */


  /* do time stats */
  if (ZERO_TIME (&ptp_save->first_time))
    {
      ptp_save->first_time = current_time;
    }
  ptp_save->last_time = current_time;



  /* bug fix:  it's legal to have the same end points reused.  The */
  /* program uses a heuristic of looking at the elapsed time from */
  /* the last packet on the previous instance and the number of FINs */
  /* in the last instance.  If we don't increment the fin_count */
  /* before bailing out in "ignore_pair" below, this heuristic breaks */

  /* figure out which direction this packet is going */
  if (dir == A2B)
    {
      thisdir = &ptp_save->a2b;
      otherdir = &ptp_save->b2a;
    }
  else
    {
      thisdir = &ptp_save->b2a;
      otherdir = &ptp_save->a2b;
    }

  /* rwnd stats */
  if (SYN_SET (ptcp))
    thisdir->rwnd_ini = th_win;
  thisdir->rwnd_tot += th_win;
  thisdir->rwnd_curr = th_win;
  if (thisdir->rwnd_max < th_win)
    thisdir->rwnd_max = th_win;


  /* meta connection stats */
  if (SYN_SET (ptcp))
    ++thisdir->syn_count;
  if (RESET_SET (ptcp))
    ++thisdir->reset_count;
  if (FIN_SET (ptcp))
    {
      ++thisdir->fin_count;
      thisdir->fin_seqno = th_seq;
    }

  /* a.c. */
  if ((ACK_SET (ptcp)) &&
      (otherdir->fin_count >= 1) && (th_ack >= (otherdir->fin_seqno + 1)))
    {
      // questo e' un ACK a un FIN => chiusa
      otherdir->closed = TRUE;
    }

  if (ACK_SET (ptcp) && otherdir->cwnd_flag)
    {
      add_histo (tcp_cwnd, (otherdir->seq - thisdir->ack));

      /* setto il flag a zero e non conto piu' 
         * il cwnd finche' non parte un altro pacchetto 
       */
      otherdir->cwnd_flag = 0;
    }

  /* a.c end */

  /* end bug fix */


  /* compute the "effective window", which is the advertised window */
  /* with scaling */
  if (ACK_SET (ptcp) || SYN_SET (ptcp))
    {
      eff_win = (u_long) th_win;

      /* N.B., the window_scale stored for the connection DURING 3way */
      /* handshaking is the REQUESTED scale.  It's only valid if both */
      /* sides request scaling.  AFTER we've seen both SYNs, that field */
      /* is reset (above) to contain zero.  Note that if we */
      /* DIDN'T see the SYNs, the windows will be off. */
      if (thisdir->f1323_ws && otherdir->f1323_ws)
	eff_win <<= thisdir->window_scale;
    }
  else
    {
      eff_win = 0;
    }


  /* idle-time stats */
  if (!ZERO_TIME (&thisdir->last_time))
    {
      u_llong itime = elapsed (thisdir->last_time, current_time);
      if (itime > thisdir->idle_max)
	thisdir->idle_max = itime;
    }
  thisdir->last_time = current_time;


  /* calculate data length */
  tcp_length = getpayloadlength (pip, plast);
  tcp_data_length = tcp_length - (4 * ptcp->th_off);

  /* congestion window 
     * questo  un pacchetto nuovo che ablita le stat su cwnd solo se ha qualcosa oltre all'ack
   */
  if ((tcp_data_length != 0) || SYN_SET (ptcp) || FIN_SET (ptcp)
      || RESET_SET (ptcp))
    {
      thisdir->cwnd_flag = 1;
    }

  /* calc. data range */
  start = th_seq;
  end = start + tcp_data_length;

  /* record sequence limits */
  if (SYN_SET (ptcp))
    {
      /* error checking - better not change! */
      if ((thisdir->syn_count > 1) && (thisdir->syn != start))
	{
	  /* it changed, that shouldn't happen! */
	  if (warn_printbad_syn_fin_seq)
	    fprintf (stderr, "rexmitted SYN had diff. seqnum! (was %lu, now %lu, etime: %d sec)\n", thisdir->syn, start, (int) (elapsed (ptp_save->first_time, current_time) / 1000000));
	  thisdir->bad_behavior = TRUE;
	}
      thisdir->syn = start;
      otherdir->ack = start;
      /* bug fix for Rob Austein <sra@epilogue.com> */
    }
  if (FIN_SET (ptcp))
    {
      /* bug fix, if there's data here too, we need to bump up the FIN */
      /* (psc data file shows example) */
      u_long fin = start + tcp_data_length;
      /* error checking - better not change! */
      if ((thisdir->fin_count > 1) && (thisdir->fin != fin))
	{
	  /* it changed, that shouldn't happen! */
	  if (warn_printbad_syn_fin_seq)
	    fprintf (stderr, "rexmitted FIN had diff. seqnum! (was %lu, now %lu, etime: %d sec)\n", thisdir->fin, fin, (int) (elapsed (ptp_save->first_time, current_time) / 1000000));
	  thisdir->bad_behavior = TRUE;
	}
      thisdir->fin = fin;
    }

  /* "ONLY" bug fix - Wed Feb 24, 1999 */
  /* the tcp-splicing heuristic needs "windowend", which was only being */
  /* calculated BELOW the "only" point below.  Move that part of the */
  /* calculation up here! */

  /* remember the OLD window end for graphing */
  /* (bug fix - Thu Aug 12, 1999) */
  old_this_windowend = thisdir->windowend;

  if (ACK_SET (ptcp))
    {
      thisdir->windowend = th_ack + eff_win;
    }
  /* end bugfix */

  /* grab the address from this packet */
  CopyAddr (&tp_in.addr_pair, pip, th_sport, th_dport);


  /* check the options */
  ptcpo = ParseOptions (ptcp, plast);
  if (ptcpo->mss != -1)
    thisdir->mss = ptcpo->mss;
  if (ptcpo->ws != -1)
    {
      thisdir->window_scale = ptcpo->ws;
      thisdir->f1323_ws = TRUE;
    }
  if (ptcpo->tsval != -1)
    {
      thisdir->f1323_ts = TRUE;
    }
  /* NOW, unless BOTH sides asked for window scaling in their SYN     */
  /* segments, we aren't using window scaling */
  if (!SYN_SET (ptcp) && ((!thisdir->f1323_ws) || (!otherdir->f1323_ws)))
    {
      thisdir->window_scale = otherdir->window_scale = 0;
    }

  /* check sacks */
  if (ptcpo->sack_req)
    {
      thisdir->fsack_req = 1;
    }
  if (ptcpo->sack_count > 0)
    {
      ++thisdir->sacks_sent;
    }

  /* do data stats */
  if (tcp_data_length > 0)
    {
      thisdir->data_pkts += 1;
      if (PUSH_SET (ptcp))
	thisdir->data_pkts_push += 1;
      thisdir->data_bytes += tcp_data_length;
      if (tcp_data_length > thisdir->max_seg_size)
	thisdir->max_seg_size = tcp_data_length;
      if ((thisdir->min_seg_size == 0) ||
	  (tcp_data_length < thisdir->min_seg_size))
	thisdir->min_seg_size = tcp_data_length;
      /* record first and last times for data (Mallman) */
      if (ZERO_TIME (&thisdir->first_data_time))
	thisdir->first_data_time = current_time;
      thisdir->last_data_time = current_time;
    }

  /* total packets stats */
  ++ptp_save->packets;
  ++thisdir->packets;

  /* set minimum seq */
  if ((thisdir->min_seq == 0) && (start != 0))
    {
      thisdir->min_seq = start;
    }
  thisdir->max_seq = end;


  /* Kevin Lahey's ECN code */
  /* only works for IPv4 */
  if (PIP_ISV4 (pip))
    {
      ecn_ce = IP_ECT (pip) && IP_CE (pip);
    }
  cwr = CWR_SET (ptcp);
  ecn_echo = ECN_ECHO_SET (ptcp);

  /* do rexmit stats */
  retrans = FALSE;
  out_order = FALSE;
  retrans_num_bytes = 0;
  if (SYN_SET (ptcp) || FIN_SET (ptcp) || tcp_data_length > 0)
    {
      int len = tcp_data_length;
      int retrans;

      if (SYN_SET (ptcp))
	++len;
      if (FIN_SET (ptcp))
	++len;

      retrans = retrans_num_bytes = rexmit (thisdir, start, len, &out_order);

      if (out_order)
	{
	  if (start == (thisdir->last_ooo_seq))
	    {
	      /* continua un burst di ooo */
	      thisdir->curr_ooo_burst += len;
	    }
	  else
	    {
	      /* chiudo il burst precedente e lo histizzo */
	      /* solo se ho qualcosa (serve a saltare il primo giro) */
	      if (thisdir->curr_ooo_burst != 0)
		{
		  if (SYN_SET (ptcp) || FIN_SET (ptcp))
		    {
		      ooo_stat_sf_add (pip, thisdir, dir);
		    }
		  else if (otherdir->rwnd_curr_burst == 0)
		    {
		      ooo_stat_rwndz_add (pip, thisdir, dir);
		    }
		  else
		    {
		      ooo_stat_add (pip, thisdir, dir);
		    }
		}

	      /* e ne apro uno nuovo */
	      thisdir->curr_ooo_burst = len;
	    }
	  /* in ogni caso aggiorno la seq dell'ooo... */
	  thisdir->last_ooo_seq = start + len;
	  /* ...memorizzo se c'e' sf... */
	  if (SYN_SET (ptcp) || FIN_SET (ptcp))
	    {
	      thisdir->curr_ooo_sf = 1;
	    }
	  else
	    {
	      thisdir->curr_ooo_sf = 0;
	    }
	  /* ...otherdir->rwnd_curr lo aggiorno solo se sono in un burst    */
	  /* serve a identificare ritrasmissioni di pacchetti per "probing" */
	  otherdir->rwnd_curr_burst = otherdir->rwnd_curr;
	}


      if (retrans != 0)
	{
	  if (start == (thisdir->last_dup_seq))
	    {
	      /* continua un burst di dup */
	      thisdir->curr_dup_burst += len;
	    }
	  else
	    {
	      /* chiudo il burst precedente e lo histizzo */
	      /* solo se ho qualcosa (serve a saltare il primo giro) */
	      if (thisdir->curr_dup_burst != 0)
		{
		  if (SYN_SET (ptcp) || FIN_SET (ptcp))
		    {
		      dup_stat_sf_add (pip, thisdir, dir);
		    }
		  else if (otherdir->rwnd_curr_burst == 0)
		    {
		      dup_stat_rwndz_add (pip, thisdir, dir);
		    }
		  else
		    {
		      dup_stat_add (pip, thisdir, dir);
		    }
		}

	      /* e ne apro uno nuovo */
	      thisdir->curr_dup_burst = len;
	    }
	  /* in ogni caso aggiorno la seq del dup... */
	  thisdir->last_dup_seq = start + len;
	  /* ..memorizzo se c'e' sf.. */
	  if (SYN_SET (ptcp) || FIN_SET (ptcp))
	    {
	      thisdir->curr_dup_sf = 1;
	    }
	  else
	    {
	      thisdir->curr_dup_sf = 0;
	    }
	  /* ...otherdir->rwnd_curr lo aggiorno solo se sono in un burst    */
	  /* serve a identificare ritrasmissioni di pacchetti per "probing" */
	  otherdir->rwnd_curr_burst = otherdir->rwnd_curr;
	}

      if (out_order)
	++thisdir->out_order_pkts;

      /* count anything NOT retransmitted as "unique" */
      /* exclude SYN and FIN */
      if (SYN_SET (ptcp))
	{
	  /* don't count the SYN as data */
	  --len;
	  /* if the SYN was rexmitted, then don't count it */
	  if (thisdir->syn_count > 1)
	    --retrans;
	}
      if (FIN_SET (ptcp))
	{
	  /* don't count the FIN as data */
	  --len;
	  /* if the FIN was rexmitted, then don't count it */
	  if (thisdir->fin_count > 1)
	    --retrans;
	}
      if (retrans < len)
	thisdir->unique_bytes += (len - retrans);
    }


  /* do rtt stats */
  if (ACK_SET (ptcp))
    {
      ack_type = ack_in (otherdir, th_ack, tcp_data_length);
    }

  /* stats for rexmitted data */
  if (retrans_num_bytes > 0)
    {
      retrans = TRUE;
      thisdir->rexmit_pkts += 1;
      thisdir->rexmit_bytes += retrans_num_bytes;
    }
  else
    {
      thisdir->seq = end;
    }

  /* check for RESET */
  if (RESET_SET (ptcp))
    {

      if (ACK_SET (ptcp))
	++thisdir->ack_pkts;

      /* .a.c. start */
      if (ConnReset (ptp_save))
	{
	  //fprintf (stdout, "  (new reset)\n");
	  tot_conn--;
	  make_conn_stats (ptp_save,
	     (ptp_save->b2a.syn_count >0 && ptp_save->a2b.syn_count>0));

	  /* rilascio la memoria delle statistiche per questo flusso */

	  free_tp (ptp_save);

	  /* rilascio l'elemento della lista appesa sotto la hash */
	  ptph_tmp = ptph_save;
	  *(ptph_save->ttp_ptr) = NULL;
	  *ptph_ptr = ptph_save->next;
	  ptph_release (ptph_tmp);
	}
      /* .a.c. end */

      return (ptp_save);
    }


  /* do window stats (include first SYN too!) */
  if (ACK_SET (ptcp) || SYN_SET (ptcp))
    {
      if (eff_win > thisdir->win_max)
	thisdir->win_max = eff_win;
      if ((eff_win > 0) &&
	  ((thisdir->win_min == 0) || (eff_win < thisdir->win_min)))
	thisdir->win_min = eff_win;
      thisdir->win_tot += eff_win;
    }

  if (ACK_SET (ptcp))
    {
      seqnum ack = th_ack;
      u_long winend;

      winend = ack + eff_win;

      if (eff_win == 0)
	++thisdir->win_zero_ct;

      ++thisdir->ack_pkts;
      if ((tcp_data_length == 0) &&
	  !SYN_SET (ptcp) && !FIN_SET (ptcp) && !RESET_SET (ptcp))
	{
	  ++thisdir->pureack_pkts;
	}

      thisdir->time = current_time;
      thisdir->ack = ack;

    }				/* end ACK_SET(ptcp) */

  /* do stats for initial window (first slow start) */
  /* (if there's data in this and we've NEVER seen */
  /*  an ACK coming back from the other side) */
  /* this is for Mark Allman for slow start testing -- Mon Mar 10, 1997 */
  if (!otherdir->data_acked && ACK_SET (ptcp)
      && ((otherdir->syn + 1) != th_ack))
    {
      otherdir->data_acked = TRUE;
    }
  if ((tcp_data_length > 0) && (!thisdir->data_acked))
    {
      if (!retrans)
	{
	  /* don't count it if it was retransmitted */
	  thisdir->initialwin_bytes += tcp_data_length;
	  thisdir->initialwin_segs += 1;
	}
    }

  if (SYN_SET (ptcp) && !ACK_SET (ptcp))
    {
      thisdir->highest_seqno = thisdir->max_seq;
      thisdir->payload_end_time = current_time;
      otherdir->payload_end_time = current_time;
    }

  /* check if this segment is carrying new or retransmitted data */
  if (thisdir->max_seq >= thisdir->highest_seqno && (tcp_data_length != 0))
    {
      thisdir->highest_seqno = thisdir->max_seq;
      thisdir->data_to_be_acked = TRUE;
    }

  /* update payload_end_time if this is a true ack */
  if (ACK_SET (ptcp) && otherdir->data_to_be_acked)
    {
      otherdir->payload_end_time = current_time;
      if (th_ack > otherdir->highest_seqno)
	otherdir->data_to_be_acked = FALSE;
    }

  /* do stats for congestion window (estimated) */
  /* estimate the congestion window as the number of outstanding */
  /* un-acked bytes */
  if (!SYN_SET (ptcp) && !out_order && !retrans)
    {
      u_long cwin = end - otherdir->ack;

      if (cwin > thisdir->cwin_max)
	thisdir->cwin_max = cwin;
      if ((cwin > 0) &&
	  ((thisdir->cwin_min == 0) || (cwin < thisdir->cwin_min)))
	thisdir->cwin_min = cwin;
    }

  /* .a.c. vediamo se la connessione  finita */

  if (ConnComplete (ptp_save))
    {
      //fprintf (stdout, "  (new complete)\n");

      tot_conn--;
      make_conn_stats (ptp_save, 1);

      /* rilascio la memoria delle statistiche per questo flusso */

      free_tp (ptp_save);

      /* rilascio l'elemento della lista appesa sotto la hash */
      ptph_tmp = ptph_save;
      *(ptph_save->ttp_ptr) = NULL;

      /* ptph_ptr  la head, quella in cima alla hash          */
      /*  Il mio elemento  in cima alla hash perch' lo porto  */
      /*  su quando lo esamno con FindTTP                      */
      *ptph_ptr = ptph_save->next;
      ptph_release (ptph_tmp);
    }


  /* .a.c. end */

  return (ptp_save);
}

void
print_ttp ()
{
  int p;

  for (p = 0; p < MAX_TCP_PAIRS; p++)
    {
      printf ("[%2d]", p);
      if (ttp[p] != NULL)
	printf ("->[ptp]\n");
      else
	printf ("->[NULL]\n");
    }
}

void
trace_done2 (void)
{
  tcp_pair *ptp;
  int ix;

  /* complete the "idle time" calculations using NOW */
  for (ix = 0; ix <= num_tcp_pairs; ++ix)
    {
      tcb *thisdir;
      u_llong itime;

      ptp = ttp[ix];

      if (ptp == NULL)		/* gia analizzato */
	continue;

/* if it's CLOSED, skip it *//* dovrebbe essere inutile */
      if ((FinCount (ptp) >= 2) || (ConnReset (ptp)))
	continue;

      /* a2b direction */
      thisdir = &ptp->a2b;
      if (!ZERO_TIME (&thisdir->last_time))
	{
	  itime = elapsed (thisdir->last_time, current_time);
	  if (itime > thisdir->idle_max)
	    thisdir->idle_max = itime;
	}


      /* b2a direction */
      thisdir = &ptp->b2a;
      if (!ZERO_TIME (&thisdir->last_time))
	{
	  itime = elapsed (thisdir->last_time, current_time);
	  if (itime > thisdir->idle_max)
	    thisdir->idle_max = itime;
	}
    }

  /* Qui, se voglio, escludo dalle statistiche le connessioni 
     che restano: sono non si son chiuse "bene"  */
  for (ix = 0; ix <= num_tcp_pairs; ++ix)
    {
      ptp = ttp[ix];
      if (ptp == NULL)		/* gia' analizzato */
	continue;

      make_conn_stats (ptp, 0);
    }
}

void
trace_done_periodic ()
{
  tcp_pair *ptp;
  int ix, dir, j;
  unsigned int cleaned = 0;
  unsigned long init_tot_conn = tot_conn;
  extern ptp_snap *ptp_hashtable[];
  hash hval;
  ptp_snap *ptph_tmp, *ptph, *ptph_prev;
  ptp_snap **pptph_head = NULL;

  /* complete the "idle time" calculations using NOW */
  if (printticks)
     fprintf (stderr, "\nStart cleaning flows");
  for (ix = 0; ix <= MAX_TCP_PAIRS; ++ix)
    {
      tcb *thisdir;
      u_llong itime;

      ptp = ttp[ix];

      if ((ptp == NULL))	/* gia analizzato */
	continue;


      /* Se ho visto passare l'ultimo pacchetto piu' di IDLE_TIME minuti fa... */
      if ((elapsed (ptp->last_time, current_time) > IDLE_TIME))
	{
	  cleaned++;

	  /* a2b direction */
	  thisdir = &ptp->a2b;
	  if (!ZERO_TIME (&thisdir->last_time))
	    {
	      itime = elapsed (thisdir->last_time, current_time);
	      if (itime > thisdir->idle_max)
		thisdir->idle_max = itime;
	    }


	  /* b2a direction */
	  thisdir = &ptp->b2a;
	  if (!ZERO_TIME (&thisdir->last_time))
	    {
	      itime = elapsed (thisdir->last_time, current_time);
	      if (itime > thisdir->idle_max)
		thisdir->idle_max = itime;
	    }

	  /* Passing 0 this flow will not be considered to make Histograms... */
	  make_conn_stats (ptp, 0);
	  tot_conn--;

	  /* rilascio l'elemento della hash ... */
	  hval = ptp->addr_pair.hash % HASH_TABLE_SIZE;

	  pptph_head = &ptp_hashtable[hval];
	  j = 0;
	  ptph_prev = *pptph_head;
	  for (ptph = *pptph_head; ptph; ptph = ptph->next)
	    {
	      j++;
	      if (SameConn (&ptp->addr_pair, &ptph->addr_pair, &dir))
		{
		  ptph_tmp = ptph;
		  if (j == 1)
		    {
		      /*  in cima alla lista */
		      ptp_hashtable[hval] = ptph->next;
		    }
		  else
		    {
		      /*  in mezzo alla lista */
		      ptph_prev->next = ptph->next;
		    }
		  ptph_release (ptph_tmp);
		  break;
		}
	      ptph_prev = ptph;
	    }

	  /* e rilascio il tcp_pair. */
	  free_tp (ptp);
	  ttp[ix] = NULL;
	}

    }
  if (printticks)
     fprintf (stderr, "\rCleaned %d/(%ld)flows \n", cleaned, init_tot_conn);
}

void
trace_init (void)
{
  static Bool initted = FALSE;

  if (initted)
    return;

  initted = TRUE;

  /* create an array to hold any pairs that we might create */
  ttp = (tcp_pair **) MallocZ (MAX_TCP_PAIRS * sizeof (tcp_pair *));

  Minit ();
}


/* get a long (4 byte) option (to avoid address alignment problems) */
static u_long
get_long_opt (void *ptr)
{
  u_long l;
  memcpy (&l, ptr, sizeof (u_long));
  return (l);
}


/* get a short (2 byte) option (to avoid address alignment problems) */
static u_short
get_short_opt (void *ptr)
{
  u_short s;
  memcpy (&s, ptr, sizeof (u_short));
  return (s);
}


struct tcp_options *
ParseOptions (struct tcphdr *ptcp, void *plast)
{
  static struct tcp_options tcpo;
  struct sack_block *psack;
  u_char *pdata;
  u_char *popt;
  u_char *plen;

  popt = (u_char *) ptcp + sizeof (struct tcphdr);
  pdata = (u_char *) ptcp + ptcp->th_off * 4;
  /* init the options structure */
  memset (&tcpo, 0, sizeof (tcpo));
  tcpo.mss = tcpo.ws = tcpo.tsval = tcpo.tsecr = -1;
  tcpo.sack_req = 0;
  tcpo.sack_count = -1;
  tcpo.echo_req = tcpo.echo_repl = -1;
  tcpo.cc = tcpo.ccnew = tcpo.ccecho = -1;

  /* a quick sanity check, the unused (MBZ) bits must BZ! */
  if (warn_printbadmbz)
    {
      if (ptcp->th_x2 != 0)
	{
	  fprintf (stderr,
		   "TCP packet %lu: 4 reserved bits are not zero (0x%01x)\n",
		   pnum, ptcp->th_x2);
	}
      if ((ptcp->th_flags & 0xc0) != 0)
	{
	  fprintf (stderr,
		   "TCP packet %lu: upper flag bits are not zero (0x%02x)\n",
		   pnum, ptcp->th_flags);
	}
    }
  else
    {
      static int warned = 0;
      if (!warned && ((ptcp->th_x2 != 0) || ((ptcp->th_flags & 0xc0) != 0)))
	{
	  warned = 1;
	  fprintf (stderr, "\
TCP packet %lu: reserved bits are not all zero.  \n\
\tFurther warnings disabled, use '-w' for more info\n", pnum);
	}
    }

  /* looks good, now check each option in turn */
  while (popt < pdata)
    {
      plen = popt + 1;
      /* check for truncation error */
      if ((unsigned long) popt >= (unsigned long) plast)
	{
	  if (warn_printtrunc)
	    fprintf (stderr, "\
ParseOptions: packet %lu too short to parse remaining options\n", pnum);
	  ++ctrunc;
	  break;
	}

#define CHECK_O_LEN(opt) \
	if (*plen == 0) { \
	    if (warn_printtrunc) fprintf(stderr, "\
ParseOptions: packet %lu %s option has length 0, skipping other options\n", \
                                           pnum,opt); \
	    popt = pdata; break;} \
	if ((unsigned)popt + *plen - 1 > (unsigned)(plast)) { \
	    if (warn_printtrunc) \
		fprintf(stderr, "\
ParseOptions: packet %lu %s option truncated, skipping other options\n", \
              pnum,opt); \
	    ++ctrunc; \
	    popt = pdata; break;} \


      switch (*popt)
	{
	case TCPOPT_EOL:
	  ++popt;
	  break;
	case TCPOPT_NOP:
	  ++popt;
	  break;
	case TCPOPT_MAXSEG:
	  CHECK_O_LEN ("TCPOPT_MAXSEG");
	  tcpo.mss = ntohs (get_short_opt (popt + 2));
	  popt += *plen;
	  break;
	case TCPOPT_WS:
	  CHECK_O_LEN ("TCPOPT_WS");
	  tcpo.ws = *((u_char *) (popt + 2));
	  popt += *plen;
	  break;
	case TCPOPT_TS:
	  CHECK_O_LEN ("TCPOPT_TS");
	  tcpo.tsval = ntohl (get_long_opt (popt + 2));
	  tcpo.tsecr = ntohl (get_long_opt (popt + 6));
	  popt += *plen;
	  break;
	case TCPOPT_ECHO:
	  CHECK_O_LEN ("TCPOPT_ECHO");
	  tcpo.echo_req = ntohl (get_long_opt (popt + 2));
	  popt += *plen;
	  break;
	case TCPOPT_ECHOREPLY:
	  CHECK_O_LEN ("TCPOPT_ECHOREPLY");
	  tcpo.echo_repl = ntohl (get_long_opt (popt + 2));
	  popt += *plen;
	  break;
	case TCPOPT_CC:
	  CHECK_O_LEN ("TCPOPT_CC");
	  tcpo.cc = ntohl (get_long_opt (popt + 2));
	  popt += *plen;
	  break;
	case TCPOPT_CCNEW:
	  CHECK_O_LEN ("TCPOPT_CCNEW");
	  tcpo.ccnew = ntohl (get_long_opt (popt + 2));
	  popt += *plen;
	  break;
	case TCPOPT_CCECHO:
	  CHECK_O_LEN ("TCPOPT_CCECHO");
	  tcpo.ccecho = ntohl (get_long_opt (popt + 2));
	  popt += *plen;
	  break;
	case TCPOPT_SACK_PERM:
	  CHECK_O_LEN ("TCPOPT_SACK_PERM");
	  tcpo.sack_req = 1;
	  popt += *plen;
	  break;
	case TCPOPT_SACK:
	  /* see which bytes are acked */
	  CHECK_O_LEN ("TCPOPT_SACK");
	  tcpo.sack_count = 0;
	  psack = (sack_block *) (popt + 2);	/* past the kind and length */
	  popt += *plen;
	  while ((unsigned) psack < (unsigned) popt)
	    {
	      struct sack_block *psack_local =
		&tcpo.sacks[(unsigned) tcpo.sack_count];
	      /* warning, possible alignment problem here, so we'll
	         use memcpy() and hope for the best */
	      /* better use -fno-builtin to avoid gcc alignment error
	         in GCC 2.7.2 */
	      memcpy (psack_local, psack, sizeof (sack_block));

	      /* convert to local byte order (Jamshid Mahdavi) */
	      psack_local->sack_left = ntohl (psack_local->sack_left);
	      psack_local->sack_right = ntohl (psack_local->sack_right);

	      ++psack;
	      if ((unsigned) psack > ((unsigned) plast + 1))
		{
		  /* this SACK block isn't all here */
		  if (warn_printtrunc)
		    fprintf (stderr, "packet %lu: SACK block truncated\n",
			     pnum);
		  ++ctrunc;
		  break;
		}
	      ++tcpo.sack_count;
	      if (tcpo.sack_count > MAX_SACKS)
		{
		  /* this isn't supposed to be able to happen */
		  fprintf (stderr,
			   "Warning, internal error, too many sacks!!\n");
		  tcpo.sack_count = MAX_SACKS;
		}
	    }
	  break;
	default:
	  if (debug)
	    fprintf (stderr,
		     "Warning, ignoring unknown TCP option 0x%x\n", *popt);
	  CHECK_O_LEN ("TCPOPT_UNKNOWN");

	  /* record it anyway... */
	  if (tcpo.unknown_count < MAX_UNKNOWN)
	    {
	      int ix = tcpo.unknown_count;	/* make lint happy */
	      tcpo.unknowns[ix].unkn_opt = *popt;
	      tcpo.unknowns[ix].unkn_len = *plen;
	    }
	  ++tcpo.unknown_count;

	  popt += *plen;
	  break;
	}
    }

  return (&tcpo);
}


/* given a tcp_pair and a packet, tell me which tcb it is */
struct tcb *
ptp2ptcb (tcp_pair * ptp, struct ip *pip, struct tcphdr *ptcp)
{
  int dir = 0;
  tcp_pair tp_in;

  /* grab the address from this packet */
  CopyAddr (&tp_in.addr_pair, pip,
	    ntohs (ptcp->th_sport), ntohs (ptcp->th_dport));

  /* check the direction */
  if (!SameConn (&tp_in.addr_pair, &ptp->addr_pair, &dir))
    return (NULL);		/* not found, internal error */

  if (dir == A2B)
    return (&ptp->a2b);
  else
    return (&ptp->b2a);
}


/*------------------------------------------------------------------------
 *  cksum  -  Return 16-bit ones complement of 16-bit ones complement sum 
 *------------------------------------------------------------------------
 */
static u_short
cksum (void *pvoid,		/* any alignment is legal */
       int nbytes)
{
  u_char *pchar = pvoid;
  u_long sum = 0;

  while (nbytes >= 2)
    {
      /* can't assume pointer alignment :-( */
      sum += (pchar[0] << 8);
      sum += pchar[1];

      pchar += 2;
      nbytes -= 2;
    }

  /* special check for odd length */
  if (nbytes == 1)
    {
      sum += (pchar[0] << 8);
      /* lower byte is assumed to be 0 */
    }

  sum = (sum >> 16) + (sum & 0xffff);	/* add in carry   */
  sum += (sum >> 16);		/* maybe one more */

  return (sum);
}

/* compute IP checksum */
static u_short
ip_cksum (struct ip *pip, void *plast)
{
  u_short sum;

#ifdef SUPPORT_IPV6
  if (PIP_ISV6 (pip))
    return (0);			/* IPv6 has no header checksum */
#endif
  if (!PIP_ISV4 (pip))
    return (1);			/* I have no idea! */


  /* quick sanity check, if the packet is truncated, pretend it's valid */
  if (plast < (void *) ((char *) pip + pip->ip_hl * 4 - 1))
    {
      return (0);
    }

  /* ... else IPv4 */
  sum = cksum (pip, pip->ip_hl * 4);
  return (sum);
}


/* is the IP checksum valid? */
Bool
ip_cksum_valid (struct ip * pip, void *plast)
{
  u_short sum;

  sum = ip_cksum (pip, plast);

  return ((sum == 0) || (sum == 0xffff));
}


/* compute the TCP checksum */
static u_short
tcp_cksum (struct ip *pip, struct tcphdr *ptcp, void *plast)
{
  u_long sum = 0;
  unsigned tcp_length;

  /* verify version */
  if (!PIP_ISV4 (pip) && !PIP_ISV6 (pip))
    {
      fprintf (stderr, "Internal error, tcp_cksum: neither IPv4 nor IPv6\n");
      exit (-1);
    }


  /* TCP checksum includes: */
  /* - IP source */
  /* - IP dest */
  /* - IP type */
  /* - TCP header length + TCP data length */
  /* - TCP header and data */

  if (PIP_ISV4 (pip))
    {
      /* quick sanity check, if the packet is fragmented,
         pretend it's valid */
      if ((ntohs (pip->ip_off) << 2) != 0)
	{
	  /* both the offset AND the MF bit must be 0 */
	  /* (but we shifted off the DF bit */
	  return (0);
	}

      /* 2 4-byte numbers, next to each other */
      sum += cksum (&pip->ip_src, 4 * 2);

      /* type */
      sum += (u_short) pip->ip_p;

      /* length (TCP header length + TCP data length) */
      tcp_length = ntohs (pip->ip_len) - (4 * pip->ip_hl);
      sum += (u_short) htons (tcp_length);
    }
#ifdef SUPPORT_IPV6
  else
    {				/* if (PIP_ISV6(pip)) */

      static Bool warned = FALSE;

      /* wow, this gets ugly with pseudo headers, sounds like a good
         job for another day :-(  */

      if (!warned)
	{
	  fprintf (stderr, "\nWarning: IPv6 TCP checksums not verified\n\n");
	  warned = TRUE;
	}
      return (0);		/* pretend it's valid */
    }
#endif

  /* quick sanity check, if the packet is truncated, pretend it's valid */
  if (plast < (void *) ((char *) ptcp + tcp_length - 1))
    {
      return (0);
    }


  /* checksum the TCP header and data */
  sum += cksum (ptcp, tcp_length);

  /* roll down into a 16-bit number */
  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);

  return (u_short) (~sum & 0xffff);
}



/* compute the UDP checksum */
static u_short
udp_cksum (struct ip *pip, struct udphdr *pudp, void *plast)
{
  u_long sum = 0;
  unsigned udp_length;

  /* WARNING -- this routine has not been extensively tested */

  /* verify version */
  if (!PIP_ISV4 (pip) && !PIP_ISV6 (pip))
    {
      fprintf (stderr, "Internal error, udp_cksum: neither IPv4 nor IPv6\n");
      exit (-1);
    }


  /* UDP checksum includes: */
  /* - IP source */
  /* - IP dest */
  /* - IP type */
  /* - UDP length field */
  /* - UDP header and data */

  if (PIP_ISV4 (pip))
    {
      /* 2 4-byte numbers, next to each other */
      sum += cksum (&pip->ip_src, 4 * 2);

      /* type */
      sum += (u_short) pip->ip_p;

      /* UDP length */
      udp_length = ntohs (pudp->uh_ulen);
      sum += pudp->uh_ulen;
    }
#ifdef SUPPORT_IPV6
  else
    {				/* if (PIP_ISV6(pip)) */

      static Bool warned = FALSE;

      /* wow, this gets ugly with pseudo headers, sounds like a good
         job for another day :-(  */

      if (!warned)
	{
	  fprintf (stderr, "\nWarning: IPv6 UDP checksums not verified\n\n");
	  warned = TRUE;
	}
      return (0);		/* pretend it's valid */
    }
#endif

  /* quick sanity check, if the packet is truncated, pretend it's valid */
  if (plast < (void *) ((char *) pudp + udp_length - 1))
    {
      return (0);
    }


  /* checksum the UDP header and data */
  sum += cksum (pudp, udp_length);

  /* roll down into a 16-bit number */
  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);

  return (u_short) (~sum & 0xffff);
}


/* is the TCP checksum valid? */
Bool
tcp_cksum_valid (struct ip * pip, struct tcphdr * ptcp, void *plast)
{
  return (tcp_cksum (pip, ptcp, plast) == 0);
}


/* is the UDP checksum valid? */
Bool
udp_cksum_valid (struct ip * pip, struct udphdr * pudp, void *plast)
{
  if (ntohs (pudp->uh_sum) == 0)
    {
      /* checksum not used */
      return (1);		/* valid */
    }

  return (udp_cksum (pip, pudp, plast) == 0);
}


/* NOTA. flusso_c = flusso completo: se la funzione viene 
 * chiamata con questo flag settato vengono aggiunte alle histo tutte le info possibili.
 * Se il flag  0 invece solo burst (ooo e dup) e viene creata e agginta la riga al log. 
 *
 */

void
make_conn_stats (tcp_pair * ptp_save, Bool flusso_c)
{
  tcb usc, ent;
  double etime;
  FILE *fp;
  char loglinea2b[500] = "";
  char loglineb2a[500] = "";
  tcb *pab, *pba;


  if (flusso_c)
    {
      fp = fp_logc;
    }
  else
    {
      fp = fp_lognc;
    }

  /* metto un paio di puntatori al tcb "entrante" e "uscente".
     *  mi serviranno parecchio 
   */

  pab = &(ptp_save->a2b);
  pba = &(ptp_save->b2a);
  if (ptp_save->internal_src)
    {
      usc = ptp_save->a2b;
      ent = ptp_save->b2a;
    }
  else if (ptp_save->internal_dst)
    {
      usc = ptp_save->b2a;
      ent = ptp_save->a2b;
    }
  else
    {
      if(warn_IN_OUT)
      {
      printf
	("WARN: This flow is neither incoming nor outgoing: src - %s; dst - %s!\n",
	 HostName (ptp_save->addr_pair.a_address), HostName (ptp_save->addr_pair.b_address));
      }
      return;
      
    }

  sprintf (loglinea2b, "%s\t%s\t%lu\t%u\t%lu\t%lu\t%lu\t%lu\t%lu\t%u\t%u\t%u\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%u\t%u\t%7f\t%u\t%d\t%d\t%lu\t%lu\t%lu\t%u",
           HostName (ptp_save->addr_pair.a_address),
           ServiceName (ptp_save->addr_pair.a_port),
	   pab->packets,
	   pab->reset_count, 
	   pab->ack_pkts, 
	   pab->pureack_pkts, 
	   pab->unique_bytes, 
	   pab->data_pkts, 
	   pab->data_bytes, 
	   pab->rexmit_pkts, 
	   pab->rexmit_bytes, 
	   pab->out_order_pkts, 
	   pab->syn_count, 
	   pab->fin_count, 
	   pab->f1323_ws, 
	   pab->f1323_ts, 
	   pab->window_scale, 
	   pab->fsack_req, 
	   pab->sacks_sent, 
	   pab->mss, 
	   pab->max_seg_size, 
	   pab->min_seg_size,
	   (int) (double)pab->data_bytes / ((double)pab->data_pkts + .001), /* avg segm size */
	   pab->win_max,
	   pab->win_min,
	   pab->win_zero_ct,
	   (pab->ack_pkts == 0 ? 0 : pab->win_tot / pab->ack_pkts),
	   pab->cwin_max,
	   pab->cwin_min,
	   pab->initialwin_bytes);

  sprintf (loglineb2a, "%s\t%s\t%lu\t%u\t%lu\t%lu\t%lu\t%lu\t%lu\t%u\t%u\t%u\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%u\t%u\t%7f\t%u\t%d\t%d\t%lu\t%lu\t%lu\t%u",
           HostName (ptp_save->addr_pair.b_address),
           ServiceName (ptp_save->addr_pair.b_port),
	   pba->packets,
	   pba->reset_count,
	   pba->ack_pkts,
           pba->pureack_pkts,
	   pba->unique_bytes,
           pba->data_pkts,
           pba->data_bytes,
	   pba->rexmit_pkts,
           pba->rexmit_bytes,
           pba->out_order_pkts,
	   pba->syn_count,
           pba->fin_count,
           pba->f1323_ws,
           pba->f1323_ts,
	   pba->window_scale,
           pba->fsack_req,
           pba->sacks_sent,
           pba->mss,
	   pba->max_seg_size,
           pba->min_seg_size,
	   (int) (double) pba->data_bytes / ((double) pba->data_pkts + .001),
	   pba->win_max,
           pba->win_min,
           pba->win_zero_ct,
	   (pba->ack_pkts == 0 ? 0 : pba->win_tot / pba->ack_pkts),
	   pba->cwin_max,
           pba->cwin_min,
           pba->initialwin_bytes);


  if (flusso_c)
    {
      /* statistiche sulle negoziazioni sack
         *  [1] A<->B     cioe' sack attivo
         *  [2] A ->B
         *  [3] A<- B
         *  [4] tutti gli altri
       */

      if (ptp_save->a2b.fsack_req && ptp_save->b2a.fsack_req)
	{
	  add_histo (tcp_opts, 1);
	}
      else if (ptp_save->a2b.fsack_req)
	{
	  add_histo (tcp_opts, 2);
	}
      else if (ptp_save->b2a.fsack_req)
	{
	  add_histo (tcp_opts, 3);
	}
      else
	{
	  add_histo (tcp_opts, 4);
	  //printf ("No sack request seen !\n");
	}

      /* statistiche sulle negoziazioni winscale 
         *  [5] A<->B     cioe' wscale attivo
         *  [6] A ->B
         *  [7] A<- B
         *  [8] tutti gli altri
       */

      if (ptp_save->a2b.f1323_ws && ptp_save->b2a.f1323_ws)
	{
	  add_histo (tcp_opts, 5);
	}
      else if (ptp_save->a2b.f1323_ws)
	{
	  add_histo (tcp_opts, 6);
	}
      else if (ptp_save->b2a.f1323_ws)
	{
	  add_histo (tcp_opts, 7);
	}
      else
	{
	  //printf ("No wscale request seen !\n");
	  add_histo (tcp_opts, 8);
	}


      /* statistiche sulle negoziazioni timestamp
         *  [9] A<->B     cioe' timestamp attivo
         *  [10] A ->B
         *  [11] A<- B
         *  [12] tutti gli altri
       */

      if (ptp_save->a2b.f1323_ts && ptp_save->b2a.f1323_ts)
	{
	  add_histo (tcp_opts, 9);
	}
      else if (ptp_save->a2b.f1323_ts)
	{
	  add_histo (tcp_opts, 10);
	}
      else if (ptp_save->b2a.f1323_ts)
	{
	  add_histo (tcp_opts, 11);
	}
      else
	{
	  //printf ("No wscale request seen !\n");
	  add_histo (tcp_opts, 12);
	}

      /* statistiche sull' mss
       */

      add_histo (tcp_mss_a, ptp_save->a2b.mss);
      add_histo (tcp_mss_b, ptp_save->b2a.mss);

      if ((ptp_save->a2b.mss == 0) || (ptp_save->b2a.mss == 0))
	{
	  add_histo (tcp_mss_used, 536);
	}
      else if (ptp_save->a2b.mss < ptp_save->b2a.mss)
	{
	  add_histo (tcp_mss_used, ptp_save->a2b.mss);
	}
      else
	{
	  add_histo (tcp_mss_used, ptp_save->b2a.mss);
	}

      /* statistiche sulla lunghezza dei flussi
         *  bytes entranti vs uscenti
       */

      add_histo (tcp_cl_b_s_usc, usc.data_bytes);
      add_histo (tcp_cl_b_s_ent, ent.data_bytes);
      add_histo (tcp_cl_b_l_usc, usc.data_bytes);
      add_histo (tcp_cl_b_l_ent, ent.data_bytes);

      /* statistiche sulla lunghezza dei flussi
         *  pacchetti entranti vs uscenti
       */

      add_histo (tcp_cl_p_usc, usc.packets);
      add_histo (tcp_cl_p_ent, ent.packets);

      /* statistiche sulla lunghezza dei flussi
         *  bytes sorgente vs destinazione
       */

      add_histo (tcp_cl_b_s_a2b, ptp_save->a2b.data_bytes);
      add_histo (tcp_cl_b_s_b2a, ptp_save->b2a.data_bytes);
      add_histo (tcp_cl_b_l_a2b, ptp_save->a2b.data_bytes);
      add_histo (tcp_cl_b_l_b2a, ptp_save->b2a.data_bytes);

      /* statistiche sulla lunghezza dei flussi
         *  pacchetti sorgente vs destinazione
       */

      add_histo (tcp_cl_p_a2b, ptp_save->a2b.packets);
      add_histo (tcp_cl_p_b2a, ptp_save->b2a.packets);
    }
  /* statistiche sulla rwnd */

  if (flusso_c)
    {
      add_histo (tcp_rwnd_ini, ptp_save->a2b.rwnd_ini);
      add_histo (tcp_rwnd_ini, ptp_save->b2a.rwnd_ini);


      ptp_save->a2b.packets == 0 ? 0 : add_histo (tcp_rwnd_avg,
						  (ptp_save->a2b.rwnd_tot /
						   ptp_save->a2b.packets));
      ptp_save->b2a.packets == 0 ? 0 : add_histo (tcp_rwnd_avg,
						  (ptp_save->b2a.rwnd_tot /
						   ptp_save->b2a.packets));

      add_histo (tcp_rwnd_max, ptp_save->a2b.rwnd_max);
      add_histo (tcp_rwnd_max, ptp_save->b2a.rwnd_max);

    }

  sprintf (loglinea2b, "%s\t%u", loglinea2b, ptp_save->a2b.rwnd_ini);
  sprintf (loglineb2a, "%s\t%u", loglineb2a, ptp_save->b2a.rwnd_ini);

  sprintf (loglinea2b, "%s\t%d", loglinea2b,
	   ptp_save->a2b.packets ==
	   0 ? 0 : (int) (ptp_save->a2b.rwnd_tot / ptp_save->a2b.packets));
  sprintf (loglineb2a, "%s\t%d", loglineb2a,
	   ptp_save->b2a.packets ==
	   0 ? 0 : (int) (ptp_save->b2a.rwnd_tot / ptp_save->b2a.packets));


  sprintf (loglinea2b, "%s\t%u", loglinea2b, ptp_save->a2b.rwnd_max);
  sprintf (loglineb2a, "%s\t%u", loglineb2a, ptp_save->b2a.rwnd_max);

  if (flusso_c)
    {
      /* statistiche sul rtt dei flussi
         *  entranti vs uscenti
       */

      /* avg */
      add_histo (rtt_avg_usc,
		 (Average (usc.rtt_sum, usc.rtt_count) / 1000.0));
      add_histo (rtt_avg_ent,
		 (Average (ent.rtt_sum, ent.rtt_count) / 1000.0));

      /* min */
      add_histo (rtt_min_usc, (usc.rtt_min / 1000.0));
      add_histo (rtt_min_ent, (ent.rtt_min / 1000.0));

      /* max */
      add_histo (rtt_max_usc, (usc.rtt_max / 1000.0));
      add_histo (rtt_max_ent, (ent.rtt_max / 1000.0));

      /* stdev */
      add_histo (rtt_stdev_usc,
		 Stdev (usc.rtt_sum, usc.rtt_sum2, usc.rtt_count) / 1000.0);
      add_histo (rtt_stdev_ent,
		 Stdev (ent.rtt_sum, ent.rtt_sum2, ent.rtt_count) / 1000.0);

      /* valid samples */
      add_histo (rtt_cnt_usc, usc.rtt_count);
      add_histo (rtt_cnt_ent, ent.rtt_count);
    }

  if (flusso_c)
    {
      /* statistiche sul rtt dei flussi
         *  sorgente vs destinazione
       */

      /* avg */
      add_histo (rtt_avg_a2b,
		 (Average (ptp_save->a2b.rtt_sum, ptp_save->a2b.rtt_count) /
		  1000.0));
      add_histo (rtt_avg_b2a,
		 (Average (ptp_save->b2a.rtt_sum, ptp_save->b2a.rtt_count) /
		  1000.0));

      /* min */
      add_histo (rtt_min_a2b, (ptp_save->a2b.rtt_min / 1000.0));
      add_histo (rtt_min_b2a, (ptp_save->b2a.rtt_min / 1000.0));

      /* max */
      add_histo (rtt_max_a2b, (ptp_save->a2b.rtt_max / 1000.0));
      add_histo (rtt_max_b2a, (ptp_save->b2a.rtt_max / 1000.0));

      /* stdev */
      add_histo (rtt_stdev_a2b,
		 Stdev (ptp_save->a2b.rtt_sum, ptp_save->a2b.rtt_sum2,
			ptp_save->a2b.rtt_count) / 1000.0);
      add_histo (rtt_stdev_b2a,
		 Stdev (ptp_save->b2a.rtt_sum, ptp_save->b2a.rtt_sum2,
			ptp_save->b2a.rtt_count) / 1000.0);

      /* valid samples */
      add_histo (rtt_cnt_a2b, ptp_save->a2b.rtt_count);
      add_histo (rtt_cnt_b2a, ptp_save->b2a.rtt_count);
    }

  sprintf (loglinea2b, "%s\t%f", loglinea2b,
	   (Average (ptp_save->a2b.rtt_sum, ptp_save->a2b.rtt_count) /
	    1000.0));
  sprintf (loglineb2a, "%s\t%f", loglineb2a,
	   (Average (ptp_save->b2a.rtt_sum, ptp_save->b2a.rtt_count) /
	    1000.0));
  sprintf (loglinea2b, "%s\t%f", loglinea2b,
	   (ptp_save->a2b.rtt_min / 1000.0));
  sprintf (loglineb2a, "%s\t%f", loglineb2a,
	   (ptp_save->b2a.rtt_min / 1000.0));
  sprintf (loglinea2b, "%s\t%f", loglinea2b,
	   (ptp_save->a2b.rtt_max / 1000.0));
  sprintf (loglineb2a, "%s\t%f", loglineb2a,
	   (ptp_save->b2a.rtt_max / 1000.0));
  sprintf (loglinea2b, "%s\t%f", loglinea2b,
	   Stdev (ptp_save->a2b.rtt_sum, ptp_save->a2b.rtt_sum2,
		  ptp_save->a2b.rtt_count) / 1000.0);
  sprintf (loglineb2a, "%s\t%f", loglineb2a,
	   Stdev (ptp_save->b2a.rtt_sum, ptp_save->b2a.rtt_sum2,
		  ptp_save->b2a.rtt_count) / 1000.0);
  sprintf (loglinea2b, "%s\t%u", loglinea2b, ptp_save->a2b.rtt_count);
  sprintf (loglineb2a, "%s\t%u", loglineb2a, ptp_save->b2a.rtt_count);

  /* ooo burst rimanente */
  /* sorgente vs destinazione */
  if (ptp_save->a2b.curr_ooo_burst != 0)
    {
      if (ptp_save->a2b.curr_ooo_sf)
	{
	  add_histo (ooob_sf_a2b, ptp_save->a2b.curr_ooo_burst);
	  ptp_save->a2b.tot_ooob_sf++;
	}
      else if (ptp_save->b2a.rwnd_curr_burst == 0)
	{
	  add_histo (ooob_rwndz_a2b, ptp_save->a2b.curr_ooo_burst);
	  ptp_save->a2b.tot_ooob_rwndz++;
	}
      else
	{
	  add_histo (ooob_a2b, ptp_save->a2b.curr_ooo_burst);
	  ptp_save->a2b.tot_ooob++;
	}
    }
  if (ptp_save->b2a.curr_ooo_burst != 0)
    {
      if (ptp_save->b2a.curr_ooo_sf)
	{
	  add_histo (ooob_sf_b2a, ptp_save->b2a.curr_ooo_burst);
	  ptp_save->b2a.tot_ooob_sf++;
	}
      else if (ptp_save->a2b.rwnd_curr_burst == 0)
	{
	  add_histo (ooob_rwndz_b2a, ptp_save->b2a.curr_ooo_burst);
	  ptp_save->b2a.tot_ooob_rwndz++;
	}
      else
	{
	  add_histo (ooob_b2a, ptp_save->b2a.curr_ooo_burst);
	  ptp_save->b2a.tot_ooob++;
	}
    }

  /* ooo burst rimanente */
  /* entrante vs. uscente */
  if (ent.curr_ooo_burst != 0)
    {
      if (ent.curr_ooo_sf)
	{
	  add_histo (ooob_sf_ent, ent.curr_ooo_burst);
	}
      else if (usc.rwnd_curr_burst == 0)
	{
	  add_histo (ooob_rwndz_ent, ent.curr_ooo_burst);
	}
      else
	{
	  add_histo (ooob_ent, ent.curr_ooo_burst);
	}
    }
  if (usc.curr_ooo_burst != 0)
    {
      if (usc.curr_ooo_sf)
	{
	  add_histo (ooob_sf_usc, usc.curr_ooo_burst);
	}
      else if (ent.rwnd_curr_burst == 0)
	{
	  add_histo (ooob_rwndz_usc, usc.curr_ooo_burst);
	}
      else
	{
	  add_histo (ooob_usc, usc.curr_ooo_burst);
	}
    }

  /* dup burst rimanente */
  /* sorgente vs destinazione */
  if (ptp_save->a2b.curr_dup_burst != 0)
    {
      if (ptp_save->a2b.curr_dup_sf)
	{
	  add_histo (dupb_sf_a2b, ptp_save->a2b.curr_dup_burst);
	  ptp_save->a2b.tot_dupb_sf++;
	}
      else if (ptp_save->b2a.rwnd_curr_burst == 0)
	{
	  add_histo (dupb_rwndz_a2b, ptp_save->a2b.curr_dup_burst);
	  ptp_save->a2b.tot_dupb_rwndz++;
	}
      else
	{
	  add_histo (dupb_a2b, ptp_save->a2b.curr_dup_burst);
	  ptp_save->a2b.tot_dupb++;
	}
    }
  if (ptp_save->b2a.curr_dup_burst != 0)
    {
      if (ptp_save->b2a.curr_dup_sf)
	{
	  add_histo (dupb_sf_b2a, ptp_save->b2a.curr_dup_burst);
	  ptp_save->a2b.tot_dupb_sf++;
	}
      else if (ptp_save->a2b.rwnd_curr_burst == 0)
	{
	  add_histo (dupb_rwndz_b2a, ptp_save->b2a.curr_dup_burst);
	  ptp_save->a2b.tot_dupb_rwndz++;
	}
      else
	{
	  add_histo (dupb_b2a, ptp_save->b2a.curr_dup_burst);
	  ptp_save->a2b.tot_dupb++;
	}
    }

  /* dup burst rimanente */
  /* entrante vs. uscente */
  if (ent.curr_dup_burst != 0)
    {
      if (ent.curr_dup_sf)
	{
	  add_histo (dupb_sf_ent, ent.curr_dup_burst);
	}
      else if (usc.rwnd_curr_burst == 0)
	{
	  add_histo (dupb_rwndz_ent, ent.curr_dup_burst);
	}
      else
	{
	  add_histo (dupb_ent, ent.curr_dup_burst);
	}
    }
  if (usc.curr_dup_burst != 0)
    {
      if (usc.curr_dup_sf)
	{
	  add_histo (dupb_sf_usc, usc.curr_dup_burst);
	}
      else if (ent.rwnd_curr_burst == 0)
	{
	  add_histo (dupb_rwndz_usc, usc.curr_dup_burst);
	}
      else
	{
	  add_histo (dupb_usc, usc.curr_dup_burst);
	}
    }

  sprintf (loglinea2b, "%s\t%lu\t%u\t%u", loglinea2b,
	   ptp_save->a2b.tot_ooob, ptp_save->a2b.tot_ooob_sf,
	   ptp_save->a2b.tot_ooob_rwndz);
  sprintf (loglineb2a, "%s\t%lu\t%u\t%u", loglineb2a,
	   ptp_save->b2a.tot_ooob, ptp_save->b2a.tot_ooob_sf,
	   ptp_save->b2a.tot_ooob_rwndz);

  sprintf (loglinea2b, "%s\t%lu\t%u\t%u", loglinea2b,
	   ptp_save->a2b.tot_dupb, ptp_save->a2b.tot_dupb_sf,
	   ptp_save->a2b.tot_dupb_rwndz);
  sprintf (loglineb2a, "%s\t%lu\t%u\t%u", loglineb2a,
	   ptp_save->b2a.tot_dupb, ptp_save->b2a.tot_dupb_sf,
	   ptp_save->b2a.tot_dupb_rwndz);

  /* Bad beahaviour*/
  sprintf (loglinea2b, "%s\t%d", loglinea2b, ptp_save->a2b.bad_behavior);
  sprintf (loglineb2a, "%s\t%d", loglineb2a, ptp_save->b2a.bad_behavior);

  /* durata totale connessione */
  /* in microsecondi: io metto in histo i millisecondi */
  etime = elapsed (ptp_save->first_time, ptp_save->last_time);
  etime = etime / 1000;
  if (flusso_c)
    add_histo (tot_time, etime);

  sprintf (loglineb2a, "%s\t%f\t", loglineb2a, etime);

  sprintf (loglineb2a, "%s\t%f\t", loglineb2a,
	   elapsed (first_packet, ptp_save->first_time) / 1000.0);
  sprintf (loglineb2a, "%s\t%f\t", loglineb2a,
	   elapsed (first_packet, ptp_save->last_time) / 1000.0);

  /* printing payload completion time */
  sprintf (loglineb2a, "%s\t%f\t", loglineb2a,
	   elapsed (first_packet, pab->payload_end_time) / 1000.0);
  sprintf (loglineb2a, "%s\t%f\t", loglineb2a,
	   elapsed (first_packet, pba->payload_end_time) / 1000.0);

  /* printing boolean flag if this is considered internal or not */
  sprintf (loglineb2a, "%s\t%d\t%d\t", loglineb2a,
	   ptp_save->internal_src, ptp_save->internal_dst);

  /* write to log file */
  fprintf (fp, "%s\t%s\n", loglinea2b, loglineb2a);
}


/* Funzioni che aggiungono alle histo per ooo e dup "standard" */

void
ooo_stat_add (struct ip *pip, tcb * thisdir, int dir)
{
#ifdef LOG_OOOB
  FILE *fp;
#endif

  if (thisdir->curr_ooo_burst <= 1000)
    {
      if (internal_src)
	{
	  add_histo (ooob_zm_usc, thisdir->curr_ooo_burst);
	}
      else
	{
	  add_histo (ooob_zm_ent, thisdir->curr_ooo_burst);
	}
    }

  if (dir == A2B)
    {
      add_histo (ooob_a2b, thisdir->curr_ooo_burst);
    }
  else
    {
      add_histo (ooob_b2a, thisdir->curr_ooo_burst);
    }
  thisdir->tot_ooob++;

  if (internal_src)
    {
      add_histo (ooob_usc, thisdir->curr_ooo_burst);
#ifdef LOG_OOOB
      if (thisdir->curr_ooo_burst <= 1000)
	{
	  fp = fopen ("ooob_zero_usc", "a");	/* linea: 2496 */
	  fprintf (fp, "%ld => %s:%s %s:%s - %lu - %u - %lu\n", pnum,
		HostName (ptp->addr_pair.a_address),
		ServiceName (ptp->addr_pair.a_port),
		HostName (ptp->addr_pair.b_address),
		ServiceName (ptp->addr_pair.b_port),
	  fclose (fp);
	}
#endif
    }
  else
    {
      add_histo (ooob_ent, thisdir->curr_ooo_burst);
#ifdef LOG_OOOB
      if (thisdir->curr_ooo_burst <= 1000)
	{
	  fp = fopen ("ooob_zero_ent", "a");	/* linea: 2496 */
	  fprintf (fp, "%ld => %s:%s %s:%s - %lu - %u - %lu\n", pnum,
		HostName (ptp->addr_pair.a_address),
		ServiceName (ptp->addr_pair.a_port),
		HostName (ptp->addr_pair.b_address),
		ServiceName (ptp->addr_pair.b_port),
		   thisdir->curr_ooo_burst, pip->ip_id, thisdir->ack);
	  fclose (fp);
	}
#endif
    }
}

void
dup_stat_add (struct ip *pip, tcb * thisdir, int dir)
{
  if (dir == A2B)
    {
      add_histo (dupb_a2b, thisdir->curr_dup_burst);
    }
  else
    {
      add_histo (dupb_b2a, thisdir->curr_dup_burst);
    }
  thisdir->tot_dupb++;

  if (internal_src)
    {
      add_histo (dupb_usc, thisdir->curr_dup_burst);
    }
  else
    {
      add_histo (dupb_ent, thisdir->curr_dup_burst);
    }
}

/* Funzioni che aggiungono alle histo per ooo e dup "SYN e FIN" */

void
ooo_stat_sf_add (struct ip *pip, tcb * thisdir, int dir)
{
  if (dir == A2B)
    {
      add_histo (ooob_sf_a2b, thisdir->curr_ooo_burst);
      thisdir->tot_ooob_sf++;
    }
  else
    {
      add_histo (ooob_sf_b2a, thisdir->curr_ooo_burst);
    }
  thisdir->tot_ooob_sf++;

  if (internal_src)
    {
      add_histo (ooob_sf_usc, thisdir->curr_ooo_burst);
    }
  else
    {
      add_histo (ooob_sf_ent, thisdir->curr_ooo_burst);
    }
}

void
dup_stat_sf_add (struct ip *pip, tcb * thisdir, int dir)
{
  if (dir == A2B)
    {
      add_histo (dupb_sf_a2b, thisdir->curr_dup_burst);
    }
  else
    {
      add_histo (dupb_sf_b2a, thisdir->curr_dup_burst);
    }
  thisdir->tot_dupb_sf++;

  if (internal_src)
    {
      add_histo (dupb_sf_usc, thisdir->curr_dup_burst);
    }
  else
    {
      add_histo (dupb_sf_ent, thisdir->curr_dup_burst);
    }
}

/* Funzioni che aggiungono alle histo per ooo e dup "rwnd zero" (rwndz)*/

void
ooo_stat_rwndz_add (struct ip *pip, tcb * thisdir, int dir)
{
  if (dir == A2B)
    {
      add_histo (ooob_rwndz_a2b, thisdir->curr_ooo_burst);
      thisdir->tot_ooob_rwndz++;
    }
  else
    {
      add_histo (ooob_rwndz_b2a, thisdir->curr_ooo_burst);
    }
  thisdir->tot_ooob_rwndz++;

  if (internal_src)
    {
      add_histo (ooob_rwndz_usc, thisdir->curr_ooo_burst);
    }
  else
    {
      add_histo (ooob_rwndz_ent, thisdir->curr_ooo_burst);
    }
}

void
dup_stat_rwndz_add (struct ip *pip, tcb * thisdir, int dir)
{
  if (dir == A2B)
    {
      add_histo (dupb_rwndz_a2b, thisdir->curr_dup_burst);
    }
  else
    {
      add_histo (dupb_rwndz_b2a, thisdir->curr_dup_burst);
    }
  thisdir->tot_dupb_rwndz++;

  if (internal_src)
    {
      add_histo (dupb_rwndz_usc, thisdir->curr_dup_burst);
    }
  else
    {
      add_histo (dupb_rwndz_ent, thisdir->curr_dup_burst);
    }
}

void
free_tp (tcp_pair * ptp_save)
{
  int i;
  /* rilascio la memoria delle statistiche per questo flusso */

  /* inizio con i quadrant che a loro volta liberano i segmenti */
  for (i = 0; i < 4; i++)
    {
      if (ptp_save->a2b.ss->pquad[i] != NULL)
	{
	  //fprintf (stdout, "  -> a2b quad %d\n",i);
	  freequad (&(ptp_save->a2b.ss->pquad[i]));
	}
    }

  for (i = 0; i < 4; i++)
    {
      if (ptp_save->b2a.ss->pquad[i] != NULL)
	{
	  //fprintf (stdout, "  -> b2a quad %d\n",i);
	  freequad (&(ptp_save->b2a.ss->pquad[i]));
	}
    }

  /* poi libero la tcp pair (quella grossa) */

  tp_release (ptp_save);
  ptp_save = NULL;

}

/* max number of letters in endpoint name */
/* (8 allows 26**8 different endpoints (209,000,000,000)
    probably plenty for now!!!!!) */
#define MAX_HOSTLETTER_LEN 8
char *
HostLetter (unsigned ix)
{
  static char name[MAX_HOSTLETTER_LEN + 1];
  static char *pname;

  /* basically, just convert to base 26 */
  pname = &name[sizeof (name) - 1];
  *pname-- = '\00';
  while (pname >= name)
    {
      unsigned digit = ix % 26;
      *pname-- = 'a' + digit;
      ix = ix / 26;
      if (ix == 0)
	return (pname + 1);
    }

  fprintf (stderr, "Fatal, too many hosts to name (max length %d)\n",
	   MAX_HOSTLETTER_LEN);
  exit (-1);
  return (NULL);		/* NOTREACHED */
}


char *
NextHostLetter (void)
{
  static int count = 0;
  return (HostLetter (count++));
}
