/*
 *
 * Copyright (c) 2001
 *	Politecnico di Torino.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code
 * distributions retain the above copyright notice and this paragraph
 * in its entirety, (2) distributions including binary code include
 * the above copyright notice and this paragraph in its entirety in
 * the documentation or other materials provided with the
 * distribution, and (3) all advertising materials mentioning features
 * or use of this software display the following acknowledgment:
 * ``This product includes software developed by the Politecnico di Torino
 * Telecomunication Networking Group.''
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Author:	Marco Mellia, Andrea Carpani
 * 		Telecomunication Networks Group
 * 		Politecnico di Torino
 * 		Torino, Italy
 *              http://http://www.tlc-networks.polito.it/index.html
 *		mellia@mail.tlc.polito.it
 *
 * Tstat is deeply based on TCPTRACE. The authors would like to thank
 * Shawn Ostermann for the development of TCPTRACE.
 * The original TCPTRACE copyright can be found in tcptrace.copyright file.
 *
*/

static char const copyright[] =
  "@(#)Copyright (c) 2001 -- Telecomunication Network Group \
     -- Politecnico di Torino.  All rights reserved.\
     Tstat is based on TCPTRACE,\
    @(#)Copyright (c) 1999 -- Shawn Ostermann -- Ohio University.\n";

#include "tstat.h"
#include "file_formats.h"
#include "version.h"
#include <time.h>

/* version information */
char *tstat_version = VERSION;


/* local routines */
static void Formats (void);
static void Help ();
static void CheckArguments (int *pargc, char *argv[]);
static void ParseArgs (char *argsource, int *pargc, char *argv[]);
static void ProcessFile (char *filename);
static void Usage (void);
static void BadArg (char *argsource, char *format, ...);
static void Version (void);

static Bool acstats = FALSE;
static Bool basedirspecified = FALSE;
static char *basenamedir;
Bool internal_src = FALSE;
Bool internal_dst = FALSE;

struct in_addr internal_net_list[MAX_INTERNAL_HOSTS];
struct in_addr internal_net_mask2[MAX_INTERNAL_HOSTS];
int internal_net_mask[MAX_INTERNAL_HOSTS];
char *internal_net_file;
int tot_internal_nets;

/* option flags and default values */
Bool printticks = FALSE;

Bool warn_IN_OUT = FALSE;
Bool warn_ooo = FALSE;
Bool warn_MAX_ = TRUE;
Bool warn_printtrunc = FALSE;
Bool warn_printbadmbz = FALSE;
Bool warn_printbadcsum = FALSE;
Bool warn_printbad_syn_fin_seq = FALSE;
Bool do_udp = TRUE;
int debug = 0;
int two_files = 1; /* used if you have two traces that store packet in
		 two different direction... works only with DAG for now.
		 */

u_long pnum = 0;
u_long ctrunc = 0;
u_long bad_ip_checksums = 0;
u_long bad_tcp_checksums = 0;
u_long bad_udp_checksums = 0;

/* globals */
int *coredump;
struct timeval current_time;

struct adx *adx_index[MAX_ADX_SLOTS];
unsigned long int fcount; /* total flow number */

/* locally global variables */
static u_long filesize = 0;
char **filenames = NULL;
int num_files = 0;
u_int numfiles;
char *cur_filename;
static char *progname;

/* for elapsed processing time */
struct timeval wallclock_start;
struct timeval wallclock_finished;
struct timeval wallclock_temp;

/* first and last packet timestamp */
timeval first_packet = { 0, 0 };
timeval last_packet = { 0, 0 };

/* .a.c. */
Bool is_stdin;
FILE *fp_logc;
FILE *fp_lognc;
#ifdef SPRINT
Bool coming_in;
#endif

static void
Help (void)
{
  fprintf (stderr,
	   "Usage: tstat  [-Nfile] [-d n] [-htuvw] [-sdir] <file1 file2> \n\n");
  fprintf (stderr, "\t-Nfile: specify the file name which contains the\n");
  fprintf (stderr, "\t        description of the internal networks.\n");
  fprintf (stderr,
	   "\t        This file must contain the subnets that will be\n");
  fprintf (stderr,
	   "\t        considered as 'internal' during the analyisis\n");
  fprintf (stderr,
	   "\t        Each subnet must be specified using IP address\n");
  fprintf (stderr,
	   "\t        on the first line and netmask on the second one:\n");
  fprintf (stderr, "\t        130.192.0.0\n");
  fprintf (stderr, "\t        255.255.0.0\n");
  fprintf (stderr, "\t        193.204.134.0\n");
  fprintf (stderr, "\t        255.255.255.0\n\n");
  fprintf (stderr, "\t-d n: activate debug level n\n");
  fprintf (stderr, "\t-h: print this help and exit\n");
  fprintf (stderr, "\t-t: print ticks showing the trace analysis progress\n");
  fprintf (stderr, "\t-u: do not trace udp packets\n");
  fprintf (stderr, "\t-v: print version and exit\n");
  fprintf (stderr, "\t-w: print [lots] of warning\n");
  fprintf (stderr, "\t-sdir: puts the trace analysis into directory\n");
  fprintf (stderr, "\t       tree dir (otherwise will be <file>.out)\n");
  fprintf (stderr, "\tfile: trace file to be analyzed\n");
  fprintf (stderr, "\t      Use 'stdin' to read from standard input.\n");
  fprintf (stderr,
	   "\t      Usefull to allow on-line capture and processing.\n");

  fprintf (stderr, "\n");
  Formats ();
  Version ();
  exit (EXIT_SUCCESS);
}



static void
BadArg (char *argsource, char *format, ...)
{
  va_list ap;

  fprintf (stderr, "\nArgument error");
  if (argsource)
    fprintf (stderr, " (from %s)", argsource);
  fprintf (stderr, ": ");

  va_start (ap, format);
  vfprintf (stderr, format, ap);
  va_end (ap);
  fprintf (stderr, "\n");

  Usage ();
}



static void
Usage (void)
{
  fprintf (stderr, "usage: %s [args...]* dumpfile [more files]*\n", progname);

  Help ();

  exit (EXIT_FAILURE);
}



static void
Version (void)
{
  fprintf (stderr, "Version: %s\n", tstat_version);
  fprintf (stderr, "  Compiled by '%s' at '%s' on machine '%s'\n",
	   built_bywhom, built_when, built_where);
}


static void
Formats (void)
{
  int i;

  fprintf (stderr, "Supported Input File Formats:\n");
  for (i = 0; i < (int) NUM_FILE_FORMATS; ++i)
    fprintf (stderr, "\t%-15s  %s\n",
	     file_formats[i].format_name, file_formats[i].format_descr);
}


int
main (int argc, char *argv[])
{
  int i,j;
  double etime;

  if (argc == 1)
    Help ();

  /* initialize internals */
  trace_init ();

  /* parse the flags */
  CheckArguments (&argc, argv);

  /* optional UDP */
  if (do_udp)
    udptrace_init ();

  /* get starting wallclock time */
  gettimeofday (&wallclock_start, NULL);

  num_files = argc;
  printf ("%d arg%s remaining, starting with '%s'\n",
	  num_files, num_files > 1 ? "s" : "", filenames[0]);

  /* knock, knock... */
  printf ("%s\n\n", VERSION);

  /* .a.c. creo tutti gli histo che mi servono */


  protocol_ent = create_histo ("protocol_ent", 0, 256, 1);
  protocol_usc = create_histo ("protocol_usc", 0, 256, 1);

  tcp_port_src_ent = create_histo ("tcp_port_src_ent", 0, 65536, 1);
  tcp_port_src_usc = create_histo ("tcp_port_src_usc", 0, 65536, 1);
  tcp_port_dst_ent = create_histo ("tcp_port_dst_ent", 0, 65536, 1);
  tcp_port_dst_usc = create_histo ("tcp_port_dst_usc", 0, 65536, 1);

  tcp_port_syn_src_ent = create_histo ("tcp_port_syn_src_ent", 0, 65536, 1);
  tcp_port_syn_src_usc = create_histo ("tcp_port_syn_src_usc", 0, 65536, 1);
  tcp_port_syn_dst_ent = create_histo ("tcp_port_syn_dst_ent", 0, 65536, 1);
  tcp_port_syn_dst_usc = create_histo ("tcp_port_syn_dst_usc", 0, 65536, 1);

  udp_port_dst_ent = create_histo ("udp_port_dst_ent", 0, 65536, 1);
  udp_port_dst_usc = create_histo ("udp_port_dst_usc", 0, 65536, 1);
  
  tcp_opts = create_histo ("tcp_opts", 1, 15, 1);
  tcp_mss_a = create_histo ("tcp_mss_a", 0, 1600, 4);
  tcp_mss_b = create_histo ("tcp_mss_b", 0, 1600, 4);
  tcp_mss_used = create_histo ("tcp_mss_used", 0, 1600, 4);

  tcp_rwnd_ini = create_histo ("tcp_rwnd_ini", 0, 65536, 256);
  tcp_rwnd_avg = create_histo ("tcp_rwnd_avg", 0, 65536, 256);
  tcp_rwnd_max = create_histo ("tcp_rwnd_max", 0, 65536, 256);

  tcp_cwnd = create_histo ("tcp_cwnd", 0, 65536, 256);

  ip_len_usc = create_histo ("ip_len_usc", 0, 1600, 4);
  ip_len_ent = create_histo ("ip_len_ent", 0, 1600, 4);
  ip_ttl_usc = create_histo ("ip_ttl_usc", 1, 257, 1);
  ip_ttl_ent = create_histo ("ip_ttl_ent", 1, 257, 1);
  ip_tos_usc = create_histo ("ip_tos_usc", 1, 257, 1);
  ip_tos_ent = create_histo ("ip_tos_ent", 1, 257, 1);

  tcp_cl_p_usc = create_histo ("tcp_cl_p_usc", 0, 1000, 1);
  tcp_cl_p_ent = create_histo ("tcp_cl_p_ent", 0, 1000, 1);

  tcp_cl_b_s_usc = create_histo ("tcp_cl_b_s_usc", 0, 50000, 50);
  tcp_cl_b_s_ent = create_histo ("tcp_cl_b_s_ent", 0, 50000, 50);
  tcp_cl_b_l_usc = create_histo ("tcp_cl_b_l_usc", 0, 50000000, 50000);
  tcp_cl_b_l_ent = create_histo ("tcp_cl_b_l_ent", 0, 50000000, 50000);

  tcp_cl_p_a2b = create_histo ("tcp_cl_p_a2b", 0, 1000, 1);
  tcp_cl_p_b2a = create_histo ("tcp_cl_p_b2a", 0, 1000, 1);

  tcp_cl_b_s_a2b = create_histo ("tcp_cl_b_s_a2b", 0, 50000, 50);
  tcp_cl_b_s_b2a = create_histo ("tcp_cl_b_s_b2a", 0, 50000, 50);
  tcp_cl_b_l_a2b = create_histo ("tcp_cl_b_l_a2b", 0, 50000000, 50000);
  tcp_cl_b_l_b2a = create_histo ("tcp_cl_b_l_b2a", 0, 50000000, 50000);

  rtt_min_usc = create_histo ("rtt_min_usc", 0, 3500, 10);
  rtt_min_ent = create_histo ("rtt_min_ent", 0, 3500, 10);
  rtt_min_a2b = create_histo ("rtt_min_a2b", 0, 3500, 10);
  rtt_min_b2a = create_histo ("rtt_min_b2a", 0, 3500, 10);

  rtt_avg_usc = create_histo ("rtt_avg_usc", 0, 3500, 10);
  rtt_avg_ent = create_histo ("rtt_avg_ent", 0, 3500, 10);
  rtt_avg_a2b = create_histo ("rtt_avg_a2b", 0, 3500, 10);
  rtt_avg_b2a = create_histo ("rtt_avg_b2a", 0, 3500, 10);

  rtt_max_usc = create_histo ("rtt_max_usc", 0, 3500, 10);
  rtt_max_ent = create_histo ("rtt_max_ent", 0, 3500, 10);
  rtt_max_a2b = create_histo ("rtt_max_a2b", 0, 3500, 10);
  rtt_max_b2a = create_histo ("rtt_max_b2a", 0, 3500, 10);

  rtt_stdev_usc = create_histo ("rtt_stdev_usc", 0, 3500, 10);
  rtt_stdev_ent = create_histo ("rtt_stdev_ent", 0, 3500, 10);
  rtt_stdev_a2b = create_histo ("rtt_stdev_a2b", 0, 3500, 10);
  rtt_stdev_b2a = create_histo ("rtt_stdev_b2a", 0, 3500, 10);

  rtt_cnt_usc = create_histo ("rtt_cnt_usc", 0, 200, 1);
  rtt_cnt_ent = create_histo ("rtt_cnt_ent", 0, 200, 1);
  rtt_cnt_a2b = create_histo ("rtt_cnt_a2b", 0, 200, 1);
  rtt_cnt_b2a = create_histo ("rtt_cnt_b2a", 0, 200, 1);

  ooob_ent = create_histo ("ooob_ent", 0, 15000, 100);
  ooob_usc = create_histo ("ooob_usc", 0, 15000, 100);
  ooob_zm_ent = create_histo ("ooob_zm_ent", 0, 1000, 1);
  ooob_zm_usc = create_histo ("ooob_zm_usc", 0, 1000, 1);
  ooob_a2b = create_histo ("ooob_a2b", 0, 15000, 100);
  ooob_b2a = create_histo ("ooob_b2a", 0, 15000, 100);

  dupb_ent = create_histo ("dupb_ent", 0, 15000, 100);
  dupb_usc = create_histo ("dupb_usc", 0, 15000, 100);
  dupb_a2b = create_histo ("dupb_a2b", 0, 15000, 100);
  dupb_b2a = create_histo ("dupb_b2a", 0, 15000, 100);

  ooob_sf_ent = create_histo ("ooob_sf_ent", 0, 15000, 100);
  ooob_sf_usc = create_histo ("ooob_sf_usc", 0, 15000, 100);
  ooob_sf_a2b = create_histo ("ooob_sf_a2b", 0, 15000, 100);
  ooob_sf_b2a = create_histo ("ooob_sf_b2a", 0, 15000, 100);

  dupb_sf_ent = create_histo ("dupb_sf_ent", 0, 15000, 100);
  dupb_sf_usc = create_histo ("dupb_sf_usc", 0, 15000, 100);
  dupb_sf_a2b = create_histo ("dupb_sf_a2b", 0, 15000, 100);
  dupb_sf_b2a = create_histo ("dupb_sf_b2a", 0, 15000, 100);

  ooob_rwndz_ent = create_histo ("ooob_rwndz_ent", 0, 15000, 100);
  ooob_rwndz_usc = create_histo ("ooob_rwndz_usc", 0, 15000, 100);
  ooob_rwndz_a2b = create_histo ("ooob_rwndz_a2b", 0, 15000, 100);
  ooob_rwndz_b2a = create_histo ("ooob_rwndz_b2a", 0, 15000, 100);

  dupb_rwndz_ent = create_histo ("dupb_rwndz_ent", 0, 15000, 100);
  dupb_rwndz_usc = create_histo ("dupb_rwndz_usc", 0, 15000, 100);
  dupb_rwndz_a2b = create_histo ("dupb_rwndz_a2b", 0, 15000, 100);
  dupb_rwndz_b2a = create_histo ("dupb_rwndz_b2a", 0, 15000, 100);

  tot_time = create_histo ("tot_time", 0, 720000, 50);

  /* read each file in turn */
  numfiles = argc;
  for (i = 0; i < argc; i++ )
    {
    for(j=0; j< two_files; j++)
    {
      if ((debug > 0) || (numfiles > 1))
	{
	  if (argc > 1)
	    printf ("Running file '%s' (%d of %d)\n", filenames[i+j], i+j + 1,
		    numfiles);
	  else
	    printf ("Running file '%s'\n", filenames[i]);
	}
    }

      /* do the real work */
      ProcessFile (filenames[i]);
      i+=j-1;
    }

  /* clean up output */
  if (printticks)
    printf ("\n");

  /* get ending wallclock time */
  gettimeofday (&wallclock_finished, NULL);

  /* general output */
  fprintf (stdout, "%lu packets seen, %lu TCP packets traced",
	   pnum, tcp_trace_count);
  if (do_udp)
    fprintf (stdout, ", %lu UDP packets traced", udp_trace_count);
  fprintf (stdout, "\n");

  /* actual tracefile times */
  etime = elapsed (first_packet, last_packet);
  fprintf (stdout, "trace %s elapsed time: %s\n",
	   (num_files == 1) ? "file" : "files", elapsed2str (etime));
  if (debug > 0)
    {
      fprintf (stdout, "\tfirst packet:  %s\n", ts2ascii (&first_packet));
      fprintf (stdout, "\tlast packet:   %s\n", ts2ascii (&last_packet));
    }

  /* close files, cleanup, and etc... */
/*
  if (do_udp)
    udptrace_done ();
*/
  exit (EXIT_SUCCESS);
}


// MGM
/* Create subdirs into which out files will be put */
void
create_new_outfiles (char *filename)
{
  char tmpstr[200];
  struct stat fbuf;
  char date[50];
  char logfile[200] = "";
  time_t t;
  if (!basedirspecified)
    {
      /* get the basename from the tracefile */
      if (is_stdin)
	{
	  basenamedir = strdup ("stdin");
	}
      else
	{
	  basenamedir = get_basename (filename);
	}
    }
  if (stat (basenamedir, &fbuf) != 0)
    {
      mkdir (basenamedir, 0775);
    }
  basename = (char *) malloc (100 * sizeof (char));
  if (is_stdin)
    {
      time (&t);
      strftime (date, 49, "%H_%M_%d_%b_%Y", localtime (&t));
    }
  else	/* I should know the first packet time at this point... */
    {
      strftime (date, 49, "%H_%M_%d_%b_%Y", localtime (&current_time.tv_sec));
    }
  sprintf (basename, "%s/%s.dump.gz.out", basenamedir, date);

  if (stat (basename, &fbuf) != -1)
    {
      /* remove the previous directory */
      sprintf (tmpstr, "rm -rf %s", basename);
      system (tmpstr);
    }
  mkdir (basename, 0775);
  sprintf (curr_data_dir, "%s/LAST", basename);

  /* Open the files for complete and uncomplete connection logging */

  sprintf (logfile, "%s/%s", basename, "log_complete");
  if (fp_logc != NULL)
    fclose (fp_logc);
  fp_logc = fopen (logfile, "w");
  if (fp_logc == NULL)
    {
      printf ("Could not open file %s\n", logfile);
    }

  sprintf (logfile, "%s/%s", basename, "log_nocomplete");
  if (fp_lognc != NULL)
    fclose (fp_lognc);
  fp_lognc = fopen (logfile, "w");
  if (fp_lognc == NULL)
    {
      printf ("Could not open file %s\n", logfile);
    }
}



static void
ProcessFile (char *filename)
{
  pread_f *ppread;
  int ret;
  struct ip *pip;
  struct tcphdr *ptcp;
  int phystype;
  void *phys;			/* physical transport header */
  tcp_pair *ptp;
  int fix;
  int len;
  int tlen;
  void *plast;
  struct stat str_stat;
  long int location = 0;
  u_long fpnum = 0;
/* used to count the opened flows...*/
  u_long pcount = 0;
  static int file_count = 0;

  /* share the current file name */
  cur_filename = filename;

  /* open the file header */
  if (CompOpenHeader (filename) == NULL)
    {
      exit (EXIT_FAILURE);
    }

  /* see how big the file is */
  is_stdin = FALSE;
  if (FileIsStdin (filename))
    {
      filesize = 1;
      is_stdin = TRUE;
    }
  else
    {
      if (stat (filename, &str_stat) != 0)
	{
	  perror ("stat");
	  exit (EXIT_FAILURE);
	}
      else
	filesize = str_stat.st_size;
    }

  /* determine which input file format it is... */
  ppread = NULL;
  if (debug > 1)
    printf ("NUM_FILE_FORMATS: %d\n", (int) NUM_FILE_FORMATS);
  for (fix = 0; fix < (int) NUM_FILE_FORMATS; ++fix)
    {
      if (debug > 0)
	fprintf (stderr, "Checking for file format '%s' (%s)\n",
		 file_formats[fix].format_name,
		 file_formats[fix].format_descr);
      rewind (stdin);
      ppread = (*file_formats[fix].test_func) ();
      if (ppread)
	{
	  if (debug > 0)
	    fprintf (stderr, "File format is '%s' (%s)\n",
		     file_formats[fix].format_name,
		     file_formats[fix].format_descr);
	  break;
	}
      else if (debug > 0)
	{
	  fprintf (stderr, "File format is NOT '%s'\n",
		   file_formats[fix].format_name);
	}
    }

  /* if we haven't found a reader, then we can't continue */
  if (ppread == NULL)
    {
      int count = 0;

      fprintf (stderr, "Unknown input file format\n");
      Formats ();

      /* check for ASCII, a common problem */
      rewind (stdin);
      while (TRUE)
	{
	  int ch;
	  if ((ch = getchar ()) == EOF)
	    break;
	  if (!isprint (ch))
	    break;
	  if (++count >= 20)
	    {
	      /* first 20 are all ASCII */
	      fprintf (stderr,
		       "\nThis looks like an ASCII input file to me.\n");
	      exit (EXIT_FAILURE);
	    }
	}
      exit (EXIT_FAILURE);
    }

  /* open the file for processing */
  if (CompOpenFile (filename) == NULL)
    {
      exit (EXIT_FAILURE);
    }

  /* how big is it.... (possibly compressed) */
  if (debug > 0)
    {
      /* print file size */
      printf ("Trace file size: %lu bytes\n", filesize);
    }
  location = 0;


  /* count the files */
  ++file_count;

  // MGM
  if (!is_stdin)
    {
      /* read the first packet, to get the timestamp of the trace */
      ret =
	(*ppread) (&current_time, &len, &tlen, &phys, &phystype, &pip,
		   &plast);
      /* and then rewind the stream... if it is not DAG format*/
//      if(ppread != pread_dag)
//         rewind (stdin);
    }
  create_new_outfiles (filename);

  tot_conn = 0;

  /* read each packet */
  while (TRUE)
    {
      /* read the next packet */
      ret =
	(*ppread) (&current_time, &len, &tlen, &phys, &phystype, &pip,
		   &plast);

      if (ret == 0)		/* EOF */
	break;

      /* quick sanity check, better be an IPv4/v6 packet */
      if (!PIP_ISV4 (pip) && !PIP_ISV6 (pip))
	{
	  static Bool warned = FALSE;

	  if (!warned)
	    {
	      fprintf (stderr, "Warning: saw at least one non-ip packet\n");
	      warned = TRUE;
	    }

	  if (debug)
	    fprintf (stderr,
		     "Skipping packet %lu, not an IPv4/v6 packet (version:%d)\n",
		     pnum, pip->ip_v);
	  continue;
	}

      /* another sanity check, only understand ETHERNET right now */
#ifndef SPRINT
      if (phystype != PHYS_ETHER)
	{
	  static int not_ether = 0;

	  ++not_ether;
	  if (not_ether < 5)
	    {
	      fprintf (stderr,
		       "Skipping packet %lu, not an ethernet packet\n", pnum);
	    }			/* else, just shut up */
	  continue;
	}
#endif

      /* decide wheater this is internal or external */
#ifndef SPRINT  /* I already knoe if this is internal or not... */
      internal_src = internal_ip (pip->ip_src);
      internal_dst = internal_ip (pip->ip_dst);
#endif
#ifdef SPRINT
      internal_src = coming_in;
      internal_dst = !coming_in;
#endif

      /* .a.c. */
      if (internal_src)
	{
	  add_histo (protocol_usc, pip->ip_p);
	}
      else
	{
	  add_histo (protocol_ent, pip->ip_p);
	}


      if (internal_src)
	{
	  add_histo (ip_len_usc, (float) ntohs (pip->ip_len));
	  add_histo (ip_ttl_usc, (float) pip->ip_ttl);
	  add_histo (ip_tos_usc, (float) pip->ip_tos);
	}
      else
	{
	  add_histo (ip_len_ent, (float) ntohs (pip->ip_len));
	  add_histo (ip_ttl_ent, (float) pip->ip_ttl);
	  add_histo (ip_tos_ent, (float) pip->ip_tos);
	}

      add_adx (&(pip->ip_dst), adx_index);

      /* update global and per-file packet counters */
      ++pnum;			/* global */
      ++fpnum;			/* local to this file */
      ++pcount;			/* counter per chiudere i pendenti */


      /* assegno il last_time_step solo al primo pacchetto */
      if (pnum == 1)
	last_time_step = last_cleaned =current_time;


      /* check for re-ordered packets */
      if (!ZERO_TIME (&last_packet))
	{
	  if (elapsed (last_packet, current_time) < 0)
	    {
	      /* out of order */
	      if ((file_count > 1) && (fpnum == 1))
		{
		  fprintf (stderr, "\
Warning, first packet in file %s comes BEFORE the last packet\n\
in the previous file.  That will likely confuse the program, please\n\
order the files in time if you have trouble\n", filename);
		}
	      else
		{
		  static int warned = 0;

		  if (warn_ooo)
		    {
		      fprintf (stderr, "\
Warning, packet %ld in file %s comes BEFORE the previous packet\n\
That will likely confuse the program, so be careful!\n", fpnum, filename);
		    }
		  else if (!warned)
		    {
		      fprintf (stderr, "\
Packets in file %s are out of order.\n\
That will likely confuse the program, so be careful!\n", filename);
		    }
		  warned = 1;
		}

	    }
	}


      /* install signal handler */
      if (fpnum == 1)
	{
	  signal (SIGINT, QuitSig);
	  signal (SIGUSR1, Usr1Sig);
	}


      /* progress counters */
      if (printticks)
	{
	  if (CompIsCompressed ())
	    location += tlen;	/* just guess... */
	  if (((fpnum < 100) && (fpnum % 10 == 0)) ||
	      ((fpnum < 1000) && (fpnum % 100 == 0)) ||
	      ((fpnum < 10000) && (fpnum % 1000 == 0)) ||
	      ((fpnum >= 10000) && (fpnum % 10000 == 0)))
	    {

	      unsigned frac;

	      if (debug)
		fprintf (stderr, "%s: ", cur_filename);
    	      fprintf (stderr, "Tp= %lu Tf=%lu ", fpnum, fcount);
	      if (CompIsCompressed ())
		{
		  frac = location / (filesize / 100);
		  if (frac <= 100)
		    fprintf (stderr, "~%u%% (compressed)", frac);
		  else
		    fprintf (stderr, "~100%% + %u%% (compressed)",
			     frac - 100);
		}
	      else if(!is_stdin)
		{
		  location = ftell (stdin);
		  frac = location / (filesize / 100);
		  fprintf (stderr, "%u%%", frac);
		}
	      /* print elapsed time */
	      {
		double etime = elapsed (first_packet, last_packet);
		fprintf (stderr, " (%s)", elapsed2str (etime));
	      }
	      /* print number of opened flow */
	      {
		fprintf (stderr, " Nf=%lu", tot_conn);
		fprintf (stderr, " Ntrash=%lu",not_id_p);
	      }

	      /* carriage return (but not newline) */
	      fprintf (stderr, "\r");
	    }
	  fflush (stderr);
	}

      /* keep track of global times */
      if (ZERO_TIME (&first_packet))
	first_packet = current_time;
      last_packet = current_time;

      /* find the start of the TCP header */
      ptcp = gettcp (pip, &plast);

      /* if that failed, it's not TCP */
      if (ptcp == NULL)
	{
	  udp_pair *pup;
	  struct udphdr *pudp;

	  /* look for a UDP header */
	  pudp = getudp (pip, &plast);

	  if (do_udp && (pudp != NULL))
	    {
	      pup = udpdotrace (pip, pudp, plast);

	    }
	  else if (pudp == NULL)
	    {
	    }
	  continue;
	}

      /* perform TCP packet analysis */

      if ((!ACK_SET (ptcp) && SYN_SET (ptcp)))
	{
	     if (internal_src)
	       {
	      add_histo (tcp_port_syn_src_usc, (float) ntohs (ptcp->th_sport));
	      add_histo (tcp_port_syn_dst_usc, (float) ntohs (ptcp->th_dport));
	       }
	       else
	       {
	      add_histo (tcp_port_syn_src_ent, (float) ntohs (ptcp->th_sport));
	      add_histo (tcp_port_syn_dst_ent, (float) ntohs (ptcp->th_dport));
	       }
	}

      if (internal_src)
	{
       add_histo (tcp_port_src_usc, (float) ntohs (ptcp->th_sport));
       add_histo (tcp_port_dst_usc, (float) ntohs (ptcp->th_dport));
	}
	else
	{
       add_histo (tcp_port_src_ent, (float) ntohs (ptcp->th_sport));
       add_histo (tcp_port_dst_ent, (float) ntohs (ptcp->th_dport));
	}

      ptp = dotrace_2 (pip, ptcp, plast);

      /* if it wasn't "interesting", we return NULL here */
      if (ptp == NULL)
	continue;

      /* for efficiency, only allow a signal every 1000 packets       */
      /* (otherwise the system call overhead will kill us)            */
      if (pnum % 1000 == 0)
	{
	  sigset_t mask;

	  sigemptyset (&mask);
	  sigaddset (&mask, SIGINT);

	  sigprocmask (SIG_UNBLOCK, &mask, NULL);
	  /* signal can happen EXACTLY HERE, when data structures are consistant */
	  sigprocmask (SIG_BLOCK, &mask, NULL);
	}
    }				/* end while == fine file */

  /* set ^C back to the default */
  /* (so we can kill the output if needed) */
  {
    sigset_t mask;

    sigemptyset (&mask);
    sigaddset (&mask, SIGINT);

    sigprocmask (SIG_UNBLOCK, &mask, NULL);
    signal (SIGINT, SIG_DFL);
  }

  if (acstats)
    {
      sprintf (curr_data_dir, "%s/LAST", basename);  
      if (debug > 1)
	printf ("DEB: writing addresses... ");
      print_adx (adx_index);

      if (debug > 1)
	printf ("DEB: writing stats for uncomplete traces... ");
      trace_done2 ();

      if (debug > 1)
	printf ("DEB: writing stats for complete traces... ");
      print_all_histo ();
    }
  /* close the input file */
  CompCloseFile (filename);
  dump_internal_stats (0);
}

void
QuitSig (int signum)
{
  printf ("%c\n\n", 7);		/* BELL */
  printf ("Terminating processing early on signal %d\n", signum);
  printf ("Partial result after processing %lu packets:\n\n\n", pnum);
  dump_internal_stats (1);
  printf("%s\n\n",curr_data_dir);
  print_adx (adx_index);
  trace_done2 ();
  print_all_histo ();
  exit (EXIT_FAILURE);
}

extern long int tcp_packet_count;
extern int search_count;
extern long int tot_adx_hash_count, tot_adx_list_count, adx_search_hash_count,
  adx_search_list_count;

#ifdef MEMDEBUG
void memory_debug();
#endif
void
Usr1Sig (int signum)
{
  printf ("%c\n\n", 7);		/* BELL */
  printf ("Got a signal USR1\n");
  dump_internal_stats (1);
#ifdef MEMDEBUG
  memory_debug();
#endif
}

void
dump_internal_stats (int err)
{
  double etime;
  gettimeofday (&wallclock_temp, NULL);
  etime = elapsed (wallclock_start, wallclock_temp);
  if (err)
    {
      fprintf (stderr, "\n\nDumping internal status variables:\n");
      fprintf (stderr, "total packet analized : %ld\n", pnum);
      fprintf (stderr, "total flows analized : %lu\n", fcount);
      fprintf (stderr,
	       "total iteration spent in the hash search routine : %d\n",
	       search_count);
      fprintf (stderr, "total analyzed TCP packet: %ld \n", tcp_packet_count);
      fprintf (stderr, "total trash TCP packet: %ld \n", not_id_p);
      fprintf (stderr, "average TCP search length: %f\n",
	       (double)search_count / (double)tcp_packet_count);
      fprintf (stderr, "Current opened flows: %ld \n", tot_conn);
      fprintf (stderr, "Current flow vector index: %d (%d)\n", num_tcp_pairs,
	       MAX_TCP_PAIRS);
      fprintf (stderr, "Total adx used in hash: %ld \n", tot_adx_hash_count);
      fprintf (stderr, "Total adx used in list: %ld \n", tot_adx_list_count);
      fprintf (stderr, "Total adx hash search: %ld\n", adx_search_hash_count);
      fprintf (stderr, "Total adx list search: %ld\n", adx_search_list_count);
      printf ("elapsed wallclock time: %s\n", elapsed2str (etime));
      printf ("%d pkts/sec analyzed\n",
	      (int) ((double) pnum / (etime / 1000000)));
      printf ("%d flows/sec analyzed\n",
	      (int) ((double) fcount / (etime / 1000000)));
    }
  else
    {

      printf ("Dumping internal status variables:\n");
      printf ("total packet analized : %ld\n", pnum);
      printf ("total flows analized : %lu\n", fcount);
      printf ("total iteration spent in the hash search routine : %d\n",
	      search_count);
      printf ("total analyzed TCP packet: %ld \n", tcp_packet_count);
      printf ("total trash TCP packet: %ld \n", not_id_p);
      printf ("average TCP search length: %ld\n",
	      search_count / tcp_packet_count);
      printf ("Current opened flows: %ld \n", tot_conn);
      printf ("Current flow vector index: %d (%d)\n", num_tcp_pairs,
	      MAX_TCP_PAIRS);
      printf ("Total adx used in hash: %ld \n", tot_adx_hash_count);
      printf ("Total adx used in list: %ld \n", tot_adx_list_count);
      printf ("Total adx hash search: %ld\n", adx_search_hash_count);
      printf ("Total adx list search: %ld\n", adx_search_list_count);
      printf ("elapsed wallclock time: %s\n", elapsed2str (etime));
      printf ("%d pkts/sec analyzed\n",
	      (int) ((double) pnum / (etime / 1000000)));
      printf ("%d flows/sec analyzed\n",
	      (int) ((double) fcount / (etime / 1000000)));
    }
}

void *
MallocZ (int nbytes)
{
  char *ptr;

  ptr = malloc (nbytes);
  if (ptr == NULL)
    {
      perror ("Malloc failed, fatal\n");
      fprintf (stderr, "\
when memory allocation fails, it's either because:\n\
1) You're out of swap space, talk to your local sysadmin about making more\n\
   (look for system commands 'swap' or 'swapon' for quick fixes)\n\
2) The amount of memory that your OS gives each process is too little\n\
   That's a system configuration issue that you'll need to discuss\n\
   with the system administrator\n\
");
      exit (EXIT_FAILURE);
    }

  memset (ptr, 0, nbytes);	/* BZERO */

  return (ptr);
}

void *
ReallocZ (void *oldptr, int obytes, int nbytes)
{
  char *ptr;

  ptr = realloc (oldptr, nbytes);
  if (ptr == NULL)
    {
      fprintf (stderr,
	       "Realloc failed (%d bytes --> %d bytes), fatal\n",
	       obytes, nbytes);
      perror ("realloc");
      exit (EXIT_FAILURE);
    }
  if (obytes < nbytes)
    {
      memset ((char *) ptr + obytes, 0, nbytes - obytes);	/* BZERO */
    }

  return (ptr);
}


/* convert a buffer to an argc,argv[] pair */
void
StringToArgv (char *buf, int *pargc, char ***pargv)
{
  char **argv;
  int nargs = 0;

  /* discard the original string, use a copy */
  buf = strdup (buf);

  /* (very pessimistically) make the argv array */
  argv = malloc (sizeof (char *) * ((strlen (buf) / 2) + 1));

  /* skip leading blanks */
  while ((*buf != '\00') && (isspace ((int) *buf)))
    {
      if (debug > 10)
	printf ("skipping isspace('%c')\n", *buf);
      ++buf;
    }

  /* break into args */
  for (nargs = 1; *buf != '\00'; ++nargs)
    {
      char *stringend;
      argv[nargs] = buf;

      /* search for separator */
      while ((*buf != '\00') && (!isspace ((int) *buf)))
	{
	  if (debug > 10)
	    printf ("'%c' (%d) is NOT a space\n", *buf, (int) *buf);
	  ++buf;
	}
      stringend = buf;

      /* skip spaces */
      while ((*buf != '\00') && (isspace ((int) *buf)))
	{
	  if (debug > 10)
	    printf ("'%c' (%d) IS a space\n", *buf, (int) *buf);
	  ++buf;
	}

      *stringend = '\00';	/* terminate the previous string */

      if (debug)
	printf ("  argv[%d] = '%s'\n", nargs, argv[nargs]);
    }

  *pargc = nargs;
  *pargv = argv;
}


static void
CheckArguments (int *pargc, char *argv[])
{

  /* remember the name of the program for errors... */
  progname = argv[0];

  ParseArgs ("command line", pargc, argv);

  /* make sure we found the files */
  if (filenames == NULL)
    {
      BadArg (NULL, "must specify at least one file name\n");
    }
}


static void
ParseArgs (char *argsource, int *pargc, char *argv[])
{
  int i;

  /* parse the args */
  for (i = 1; i < *pargc; ++i)
    {
      if (argv[i] == NULL)
	continue;

      if (*argv[i] == '-')
	{
	  if (argv[i][1] == '\00')	/* just a '-' */
	    Usage ();

	  while (*(++argv[i]))
	    switch (*argv[i])
	      {
	      case 'N':
		if (*(argv[i] + 1))
		  {
		    /* -Nfile */
		    internal_net_file = strdup (argv[i] + 1);
		    *(argv[i] + 1) = '\00';
		    if (!LoadInternalNets (internal_net_file))
		      {
			fprintf (stderr,
				 "Error while loading configuration\n");
			fprintf (stderr, "Could not open %s\n",
				 internal_net_file);
			exit (1);
		      }
		    acstats = TRUE;
		  }
		else
		  {
		    /* maybe -N file */
		    BadArg (argsource, "-Nfile requires a file name\n");
		  }
		break;
	      case 'd':
		++debug;
		break;
	      case 'h':
		Help ();
		break;
	      case 't':
		printticks = TRUE;
		break;
	      case 'u':
		do_udp = FALSE;
		break;
	      case 's':
		if (*(argv[i] + 1))
		  {
		    /* -sdir */
		    basenamedir = strdup (argv[i] + 1);
		    *(argv[i] + 1) = '\00';
		    basedirspecified = TRUE;
		    if (debug > 1)
		      printf ("basenamedir set to %s\n", basenamedir);
		  }
		else
		  {
		    /* maybe -s file */
		    BadArg (argsource, "-sdir requires a file name\n");
		  }
		break;
	      case 'v':
		Version ();
		exit (EXIT_SUCCESS);
		break;
	      case 'w':
		warn_printtrunc = TRUE;
		warn_printbadmbz = TRUE;
		warn_printbadcsum = TRUE;
		warn_printbad_syn_fin_seq = TRUE;
		warn_ooo = TRUE;
		break;
	      case '2':
		two_files = 2;
		break;
	      default:
		BadArg (argsource, "option '%c' not understood\n", *argv[i]);
	      }
	}
      else
	{
	  filenames = &argv[i];
	  *pargc -= i;
	  return;
	}
    }

  return;
}


int
LoadInternalNets (char *file)
{
  int i;
  FILE *fp;
  char c[50];

  fp = fopen (file, "r");
  if (fp == NULL)
    {
      return 0;
    }

  i = 0;

  while (fgets (c, 20, fp) != NULL)
    {
      inet_aton (c, &(internal_net_list[i]));
      if (fgets (c, 20, fp) != NULL)
	{
	  inet_aton (c, &(internal_net_mask2[i]));
	  internal_net_mask[i] = inet_addr (c);
	}
      else
	{
	  printf ("Cannot parse mask for net %d\n", (i + 1));
	  return 0;
	}
      if (debug)
	{
	  printf ("Adding: %s as internal net ",
		  inet_ntoa (internal_net_list[i]));
	  printf ("with mask %s (%d)\n", inet_ntoa (internal_net_mask2[i]),
		  internal_net_mask[i]);
	}
      i++;
    }
  tot_internal_nets = i;
  return 1;
}



/* the memcpy() function that gcc likes to stuff into the program has alignment
   problems, so here's MY version.  It's only used for small stuff, so the
   copy should be "cheap", but we can't be too fancy due to alignment boo boos */
void *
MemCpy (void *vp1, void *vp2, size_t n)
{
  char *p1 = vp1;
  char *p2 = vp2;

  while (n-- > 0)
    *p1++ = *p2++;

  return (vp1);
}


/* funzione: internal_net(
 * Serve per decidere se in ip  "interno" o no
 */

Bool
internal_ip (struct in_addr adx)
{
  int i;

  //printf("Checking %s \n",inet_ntoa(adx));
  for (i = 0; i < tot_internal_nets; i++)
    {
      //printf(" Against: %s \n",inet_ntoa(internal_net_list[i]));
      if ((adx.s_addr & internal_net_mask[i]) == internal_net_list[i].s_addr)
	{
	  //printf("Internal: %s\n",inet_ntoa(adx));
	  return 1;
	}
    }
  //printf("External: %s\n",inet_ntoa(adx));
  return 0;
}

/* funzione: internal_net(
 * Serve per decidere se in ip  "interno" o no.
 * confronta le notazioni numeri+punti
 */

Bool
internal_ip_string (char *adx)
{
  int i;

  for (i = 0; i < tot_internal_nets; i++)
    {
      if ((inet_addr (adx) & internal_net_mask[i]) ==
	  internal_net_list[i].s_addr)
	{
	  return 1;
	}
    }
  return 0;
}
