/*
 *
 * Copyright (c) 2001
 *	Politecnico di Torino.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code
 * distributions retain the above copyright notice and this paragraph
 * in its entirety, (2) distributions including binary code include
 * the above copyright notice and this paragraph in its entirety in
 * the documentation or other materials provided with the
 * distribution, and (3) all advertising materials mentioning features
 * or use of this software display the following acknowledgment:
 * ``This product includes software developed by the Politecnico di Torino
 * Telecomunication Networking Group.''
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Author:	Marco Mellia, Andrea Carpani
 * 		Telecomunication Networks Group
 * 		Politecnico di Torino
 * 		Torino, Italy
 *              http://http://www.tlc-networks.polito.it/index.html
 *		mellia@mail.tlc.polito.it
 *
 * Tstat is deeply based on TCPTRACE. The authors would like to thank
 * Shawn Ostermann for the development of TCPTRACE.
 * The original TCPTRACE copyright can be found in tcptrace.copyright file.
 *
*/


#include "tstat.h"

/* locally global variables */
static int packet_count = 0;
static int search_count = 0;



/* provided globals  */
int num_udp_pairs = -1;		/* how many pairs we've allocated */
udp_pair **utp = NULL;		/* array of pointers to allocated pairs */
int max_udp_pairs = 64;		/* initial value, automatically increases */
u_long udp_trace_count = 0;


/* local routine definitions */
static udp_pair *NewUTP (struct ip *, struct udphdr *);
static udp_pair *FindUTP (struct ip *, struct udphdr *, int *);
static void MoreUdpPairs (int num_needed);




static udp_pair *
NewUTP (struct ip *pip, struct udphdr *pudp)
{
  udp_pair *pup;

  /* make a new one, if possible */
  if ((num_udp_pairs + 1) >= max_udp_pairs)
    {
      MoreUdpPairs (num_udp_pairs + 1);
    }

  /* create a new UDP pair record and remember where you put it */
  ++num_udp_pairs;
  pup = utp[num_udp_pairs] = MallocZ (sizeof (udp_pair));


  /* grab the address from this packet */
  CopyAddr (&pup->addr_pair,
	    pip, ntohs (pudp->uh_sport), ntohs (pudp->uh_dport));

  /* data structure setup */
  pup->a2b.pup = pup;
  pup->b2a.pup = pup;
  pup->a2b.ptwin = &pup->b2a;
  pup->b2a.ptwin = &pup->a2b;

  return (pup);
}



/* connection records are stored in a hash table.  Buckets are linked	*/
/* lists sorted by most recent access.					*/
static udp_pair *
FindUTP (struct ip *pip, struct udphdr *pudp, int *pdir)
{
  static udp_pair *pup_hashtable[HASH_TABLE_SIZE] = { NULL };
  udp_pair **ppup_head = NULL;
  udp_pair *pup;
  udp_pair *pup_last;
  udp_pair tp_in;
  int dir;
  hash hval;

  /* grab the address from this packet */
  CopyAddr (&tp_in.addr_pair, pip,
	    ntohs (pudp->uh_sport), ntohs (pudp->uh_dport));

  /* grab the hash value (already computed by CopyAddr) */
  hval = tp_in.addr_pair.hash % HASH_TABLE_SIZE;


  pup_last = NULL;
  ppup_head = &pup_hashtable[hval];
  for (pup = *ppup_head; pup; pup = pup->next)
    {
      ++search_count;
      if (SameConn (&tp_in.addr_pair, &pup->addr_pair, &dir))
	{
	  /* move to head of access list (unless already there) */
	  if (pup != *ppup_head)
	    {
	      pup_last->next = pup->next;	/* unlink */
	      pup->next = *ppup_head;	/* move to head */
	      *ppup_head = pup;
	    }
	  *pdir = dir;
	  return (pup);
	}
      pup_last = pup;
    }

  /* Didn't find it, make a new one, if possible */
  pup = NewUTP (pip, pudp);

  /* put at the head of the access list */
  if (pup)
    {
      pup->next = *ppup_head;
      *ppup_head = pup;
    }

  *pdir = A2B;
  return (pup);
}



udp_pair *
udpdotrace (struct ip * pip, struct udphdr * pudp, void *plast)
{
  udp_pair *pup_save;
  ucb *thisdir;
  ucb *otherdir;
  udp_pair tp_in;
  int dir;
  u_short uh_sport;		/* source port */
  u_short uh_dport;		/* destination port */
  u_short uh_ulen;		/* data length */

  /* make sure we have enough of the packet */
  if ((unsigned) pudp + sizeof (struct udphdr) - 1 > (unsigned) plast)
    {
      if (warn_printtrunc)
	fprintf (stderr,
		 "UDP packet %lu truncated too short to trace, ignored\n",
		 pnum);
      ++ctrunc;
      return (NULL);
    }


  /* convert interesting fields to local byte order */
  uh_sport = ntohs (pudp->uh_sport);
  uh_dport = ntohs (pudp->uh_dport);
  uh_ulen = ntohs (pudp->uh_ulen);
  
  if (internal_src)
    {
  add_histo (udp_port_dst_ent, (float) (uh_dport));
  }
  else
  {
  add_histo (udp_port_dst_usc, (float) (uh_dport));
  }
/* stop at this level of analysis */
  ++udp_trace_count;

  return NULL;
#ifdef DOESTHISWORK
  /* make sure this is one of the connections we want */
  pup_save = FindUTP (pip, pudp, &dir);

  ++packet_count;

  if (pup_save == NULL)
    {
      return (NULL);
    }

  /* do time stats */
  if (ZERO_TIME (&pup_save->first_time))
    {
      pup_save->first_time = current_time;
    }
  pup_save->last_time = current_time;

  /* grab the address from this packet */
  CopyAddr (&tp_in.addr_pair, pip, uh_sport, uh_dport);

  /* figure out which direction this packet is going */
  if (dir == A2B)
    {
      thisdir = &pup_save->a2b;
      otherdir = &pup_save->b2a;
    }
  else
    {
      thisdir = &pup_save->b2a;
      otherdir = &pup_save->a2b;
    }


  /* do data stats */
  thisdir->packets += 1;
  thisdir->data_bytes += uh_ulen;


  /* total packets stats */
  ++pup_save->packets;
  ++thisdir->packets;

  return (pup_save);
#endif
}



void
udptrace_done (void)
{
  double etime;

  if (udp_trace_count == 0)
    {
      fprintf (stdout, "no traced UDP packets\n");
      return;
    }
  else
    {
      if ((tcp_trace_count > 0))
	printf
	  ("\n============================================================\n");
      fprintf (stdout, "UDP connection info:\n");
    }

  fprintf (stdout, "%d UDP %s traced:\n",
	   num_udp_pairs + 1,
	   num_udp_pairs == 0 ? "connection" : "connections");

  /* elapsed time */
  etime = elapsed (first_packet, last_packet);

  if (ctrunc > 0)
    {
      fprintf (stdout,
	       "*** %lu packets were too short to process at some point\n",
	       ctrunc);
      if (!warn_printtrunc)
	fprintf (stdout, "\t(use -w option to show details)\n");
    }
  if (debug > 1)
    fprintf (stdout, "average search length: %d\n",
	     search_count / packet_count);

}

static void
MoreUdpPairs (int num_needed)
{
  int new_max_udp_pairs;

  if (num_needed < max_udp_pairs)
    return;

  new_max_udp_pairs = max_udp_pairs * 4;
  while (new_max_udp_pairs < num_needed)
    new_max_udp_pairs *= 4;

  if (debug)
    printf ("trace: making more space for %d total UDP pairs\n",
	    new_max_udp_pairs);

  /* enlarge array to hold any pairs that we might create */
  utp = ReallocZ (utp,
		  max_udp_pairs * sizeof (udp_pair *),
		  new_max_udp_pairs * sizeof (udp_pair *));

  max_udp_pairs = new_max_udp_pairs;
}


void
udptrace_init (void)
{
  static Bool initted = FALSE;

  if (initted)
    return;

  initted = TRUE;

  /* create an array to hold any pairs that we might create */
/*  utp = (udp_pair **) MallocZ (max_udp_pairs * sizeof (udp_pair *));
*/
}
